@@ -57,6 +57,7 @@
 //! extern crate miniscript;
 //!
 //! use std::str::FromStr;
+//! use miniscript::NullCtx;
 //!
 //! fn main() {
 //!     let desc = miniscript::Descriptor::<
@@ -64,24 +65,45 @@
 //!     >::from_str("\
 //!         sh(wsh(or_d(\
 //!             c:pk_k(020e0338c96a8870479f2396c373cc7696ba124e8635d41b0ea581112b67817261),\
-//!             c:pk_k(020e0338c96a8870479f2396c373cc7696ba124e8635d41b0ea581112b67817261)\
+//!             c:pk_k(0250863ad64a87ae8a2fe83c1af1a8403cb53f53e486d8511dad8a04887e5b2352)\
 //!         )))\
 //!     ").unwrap();
 //!
+//!     // Sometimes it is necessary to have additional information to get the bitcoin::PublicKey
+//!     // from the MiniscriptKey which can supplied by `to_pk_ctx` parameter. For example,
+//!     // when calculating the script pubkey of a descriptor with xpubs, the secp context and
+//!     // child information maybe required.
+//!
 //!     // Derive the P2SH address
 //!     assert_eq!(
-//!         desc.address(bitcoin::Network::Bitcoin).unwrap().to_string(),
-//!         "32aAVauGwencZwisuvd3anhhhQhNZQPyHv"
+//!         desc.address(bitcoin::Network::Bitcoin, NullCtx).unwrap().to_string(),
+//!         "3CJxbQBfWAe1ZkKiGQNEYrioV73ZwvBWns"
 //!     );
 //!
+//!     // Check whether the descriptor is safe
+//!     // This checks whether all spend paths are accessible in bitcoin network.
+//!     // It maybe possible that some of the spend require more than 100 elements in Wsh scripts
+//!     // Or they contain a combination of timelock and heightlock.
+//!     assert!(desc.sanity_check().is_ok());
+//!
 //!     // Estimate the satisfaction cost
-//!     assert_eq!(desc.max_satisfaction_weight(), 293);
+//!     assert_eq!(desc.max_satisfaction_weight(NullCtx).unwrap(), 293);
 //! }
 //! ```
 //!
 //!
 #![allow(bare_trait_objects)]
 #![cfg_attr(all(test, feature = "unstable"), feature(test))]
+// Coding conventions
+#![deny(unsafe_code)]
+#![deny(non_upper_case_globals)]
+#![deny(non_camel_case_types)]
+#![deny(non_snake_case)]
+#![deny(unused_mut)]
+#![deny(dead_code)]
+#![deny(unused_imports)]
+#![deny(missing_docs)]
+
 pub extern crate bitcoin;
 #[cfg(feature = "serde")]
 pub extern crate serde;
@@ -93,18 +115,22 @@ mod macros;
 
 pub mod descriptor;
 pub mod expression;
+pub mod interpreter;
 pub mod miniscript;
 pub mod policy;
 pub mod psbt;
 
+mod util;
+
 use std::str::FromStr;
 use std::{error, fmt, hash, str};
 
 use bitcoin::blockdata::{opcodes, script};
 use bitcoin::hashes::{hash160, sha256, Hash};
 
-pub use descriptor::{Descriptor, SatisfiedConstraints};
-pub use miniscript::context::{Legacy, ScriptContext, Segwitv0};
+pub use descriptor::{Descriptor, DescriptorPublicKey, DescriptorPublicKeyCtx};
+pub use interpreter::Interpreter;
+pub use miniscript::context::{Bare, Legacy, ScriptContext, Segwitv0};
 pub use miniscript::decode::Terminal;
 pub use miniscript::satisfy::{BitcoinSig, Satisfier};
 pub use miniscript::Miniscript;
@@ -150,14 +176,25 @@ impl MiniscriptKey for String {
 }
 
 /// Trait describing public key types which can be converted to bitcoin pubkeys
-pub trait ToPublicKey: MiniscriptKey {
+/// The trait relies on Copy trait because in all practical usecases `ToPkCtx`
+/// should contain references to objects which should be cheap to `Copy`.
+// Why is this a generic instead of associated type?
+// We would like to support implementation of `ToPublicKey` for both
+// secp verification_only and secp_all. But as of rust 1.29(MSRV), the
+// feature for associated traits is unstable.
+pub trait ToPublicKey<ToPkCtx: Copy>: MiniscriptKey {
     /// Converts an object to a public key
-    fn to_public_key(&self) -> bitcoin::PublicKey;
+    /// C represents additional context information that maybe
+    /// required for deriving a bitcoin::PublicKey from MiniscriptKey
+    /// You may require secp context for crypto operations
+    /// or additional information for substituting the wildcard in
+    /// extended pubkeys
+    fn to_public_key(&self, to_pk_ctx: ToPkCtx) -> bitcoin::PublicKey;
 
     /// Computes the size of a public key when serialized in a script,
     /// including the length bytes
-    fn serialized_len(&self) -> usize {
-        if self.to_public_key().compressed {
+    fn serialized_len(&self, to_pk_ctx: ToPkCtx) -> usize {
+        if self.to_public_key(to_pk_ctx).compressed {
             34
         } else {
             66
@@ -170,15 +207,19 @@ pub trait ToPublicKey: MiniscriptKey {
     /// that calling `MiniscriptKey::to_pubkeyhash` followed by this function
     /// should give the same result as calling `to_public_key` and hashing
     /// the result directly.
-    fn hash_to_hash160(hash: &<Self as MiniscriptKey>::Hash) -> hash160::Hash;
+    fn hash_to_hash160(hash: &<Self as MiniscriptKey>::Hash, to_pk_ctx: ToPkCtx) -> hash160::Hash;
 }
 
-impl ToPublicKey for bitcoin::PublicKey {
-    fn to_public_key(&self) -> bitcoin::PublicKey {
+/// Dummy Context for impl for ToPublicKey for bitcoin::PublicKey
+#[derive(Debug, Clone, Copy)]
+pub struct NullCtx;
+
+impl ToPublicKey<NullCtx> for bitcoin::PublicKey {
+    fn to_public_key(&self, _ctx: NullCtx) -> bitcoin::PublicKey {
         *self
     }
 
-    fn hash_to_hash160(hash: &hash160::Hash) -> hash160::Hash {
+    fn hash_to_hash160(hash: &hash160::Hash, _ctx: NullCtx) -> hash160::Hash {
         *hash
     }
 }
@@ -218,15 +259,15 @@ impl fmt::Display for DummyKey {
     }
 }
 
-impl ToPublicKey for DummyKey {
-    fn to_public_key(&self) -> bitcoin::PublicKey {
+impl ToPublicKey<NullCtx> for DummyKey {
+    fn to_public_key(&self, _ctx: NullCtx) -> bitcoin::PublicKey {
         bitcoin::PublicKey::from_str(
             "0250863ad64a87ae8a2fe83c1af1a8403cb53f53e486d8511dad8a04887e5b2352",
         )
         .unwrap()
     }
 
-    fn hash_to_hash160(_: &DummyKeyHash) -> hash160::Hash {
+    fn hash_to_hash160(_: &DummyKeyHash, _ctx: NullCtx) -> hash160::Hash {
         hash160::Hash::from_str("f54a5851e9372b87810a8e60cdd2e7cfd80b6e31").unwrap()
     }
 }
@@ -313,40 +354,28 @@ pub enum Error {
     CouldNotSatisfy,
     /// Typechecking failed
     TypeCheck(String),
-    ///General error in creating descriptor
+    /// General error in creating descriptor
     BadDescriptor,
-    ///Forward-secp related errors
+    /// Forward-secp related errors
     Secp(bitcoin::secp256k1::Error),
     #[cfg(feature = "compiler")]
-    ///Compiler related errors
+    /// Compiler related errors
     CompilerError(policy::compiler::CompilerError),
-    ///Errors related to policy
+    /// Errors related to policy
     PolicyError(policy::concrete::PolicyError),
-    ///Interpreter related errors
-    InterpreterError(descriptor::InterpreterError),
+    /// Errors related to lifting
+    LiftError(policy::LiftError),
     /// Forward script context related errors
     ContextError(miniscript::context::ScriptContextError),
-    /// Bad Script Sig. As per standardness rules, only pushes are allowed in
-    /// scriptSig. This error is invoked when op_codes are pushed onto the stack
-    /// As per the current implementation, pushing an integer apart from 0 or 1
-    /// will also trigger this. This is because, Miniscript only expects push
-    /// bytes for pk, sig, preimage etc or 1 or 0 for `StackElement::Satisfied`
-    /// or `StackElement::Dissatisfied`
-    BadScriptSig,
-    ///Witness must be empty for pre-segwit transactions
-    NonEmptyWitness,
-    ///ScriptSig must be empty for pure segwit transactions
-    NonEmptyScriptSig,
-    ///Incorrect Script pubkey Hash for the descriptor. This is used for both
-    /// `PkH` and `Wpkh` descriptors
-    IncorrectPubkeyHash,
-    ///Incorrect Script pubkey Hash for the descriptor. This is used for both
-    /// `Sh` and `Wsh` descriptors
-    IncorrectScriptHash,
     /// Recursion depth exceeded when parsing policy/miniscript from string
     MaxRecursiveDepthExceeded,
     /// Script size too large
     ScriptSizeTooLarge,
+    /// Anything but c:pk(key) (P2PK), c:pk_h(key) (P2PKH), and thresh_m(k,...)
+    /// up to n=3 is invalid by standardness (bare)
+    NonStandardBareScript,
+    /// Analysis Error
+    AnalysisError(miniscript::analyzable::AnalysisError),
 }
 
 #[doc(hidden)]
@@ -360,6 +389,13 @@ where
     }
 }
 
+#[doc(hidden)]
+impl From<policy::LiftError> for Error {
+    fn from(e: policy::LiftError) -> Error {
+        Error::LiftError(e)
+    }
+}
+
 #[doc(hidden)]
 impl From<miniscript::context::ScriptContextError> for Error {
     fn from(e: miniscript::context::ScriptContextError) -> Error {
@@ -367,6 +403,13 @@ impl From<miniscript::context::ScriptContextError> for Error {
     }
 }
 
+#[doc(hidden)]
+impl From<miniscript::analyzable::AnalysisError> for Error {
+    fn from(e: miniscript::analyzable::AnalysisError) -> Error {
+        Error::AnalysisError(e)
+    }
+}
+
 #[doc(hidden)]
 impl From<bitcoin::secp256k1::Error> for Error {
     fn from(e: bitcoin::secp256k1::Error) -> Error {
@@ -385,10 +428,6 @@ impl error::Error for Error {
             _ => None,
         }
     }
-
-    fn description(&self) -> &str {
-        ""
-    }
 }
 
 // https://github.com/sipa/miniscript/pull/5 for discussion on this number
@@ -434,23 +473,14 @@ impl fmt::Display for Error {
             Error::TypeCheck(ref e) => write!(f, "typecheck: {}", e),
             Error::BadDescriptor => f.write_str("could not create a descriptor"),
             Error::Secp(ref e) => fmt::Display::fmt(e, f),
-            Error::InterpreterError(ref e) => fmt::Display::fmt(e, f),
             Error::ContextError(ref e) => fmt::Display::fmt(e, f),
             #[cfg(feature = "compiler")]
             Error::CompilerError(ref e) => fmt::Display::fmt(e, f),
             Error::PolicyError(ref e) => fmt::Display::fmt(e, f),
-            Error::BadScriptSig => f.write_str("Script sig must only consist of pushes"),
-            Error::NonEmptyWitness => f.write_str("Non empty witness for Pk/Pkh"),
-            Error::NonEmptyScriptSig => f.write_str("Non empty script sig for segwit spend"),
-            Error::IncorrectScriptHash => {
-                f.write_str("Incorrect script hash for redeem script sh/wsh")
-            }
-            Error::IncorrectPubkeyHash => {
-                f.write_str("Incorrect pubkey hash for given descriptor pkh/wpkh")
-            }
+            Error::LiftError(ref e) => fmt::Display::fmt(e, f),
             Error::MaxRecursiveDepthExceeded => write!(
                 f,
-                "Recusive depth over {} not permitted",
+                "Recursive depth over {} not permitted",
                 MAX_RECURSION_DEPTH
             ),
             Error::ScriptSizeTooLarge => write!(
@@ -458,6 +488,13 @@ impl fmt::Display for Error {
                 "Standardness rules imply bitcoin than {} bytes",
                 MAX_SCRIPT_SIZE
             ),
+            Error::NonStandardBareScript => write!(
+                f,
+                "Anything but c:pk(key) (P2PK), c:pk_h(key) (P2PKH), and thresh_m(k,...) \
+                up to n=3 is invalid by standardness (bare).
+                "
+            ),
+            Error::AnalysisError(ref e) => e.fmt(f),
         }
     }
 }
@@ -477,13 +514,6 @@ impl From<policy::concrete::PolicyError> for Error {
     }
 }
 
-#[doc(hidden)]
-impl From<descriptor::InterpreterError> for Error {
-    fn from(e: descriptor::InterpreterError) -> Error {
-        Error::InterpreterError(e)
-    }
-}
-
 /// The size of an encoding of a number in Script
 pub fn script_num_size(n: usize) -> usize {
     match n {
@@ -496,6 +526,23 @@ pub fn script_num_size(n: usize) -> usize {
     }
 }
 
+/// Returns the size of the smallest push opcode used to push a given number of bytes onto the stack
+///
+/// For sizes ≤ 75, there are dedicated single-byte opcodes, so the push size is one. Otherwise,
+/// if the size can fit into 1, 2 or 4 bytes, we use the `PUSHDATA{1,2,4}` opcode respectively,
+/// followed by the actual size encoded in that many bytes.
+fn push_opcode_size(script_size: usize) -> usize {
+    if script_size < 76 {
+        1
+    } else if script_size < 0x100 {
+        2
+    } else if script_size < 0x10000 {
+        3
+    } else {
+        5
+    }
+}
+
 /// Helper function used by tests
 #[cfg(test)]
 fn hex_script(s: &str) -> bitcoin::Script {
