@@ -19,7 +19,7 @@
 //! encoding in Bitcoin script, as well as a datatype. Full details
 //! are given on the Miniscript website.
 
-use std::{cmp, fmt, str};
+use std::{fmt, str};
 
 use bitcoin::blockdata::{opcodes, script};
 use bitcoin::hashes::hex::FromHex;
@@ -149,10 +149,6 @@ impl<Pk: MiniscriptKey, Ctx: ScriptContext> Terminal<Pk, Ctx> {
                 Terminal::Multi(k, keys?)
             }
         };
-        Ctx::check_frag_validity(&frag).expect(
-            "Translated fragment not valid.\n
-        Uncompressed Pubkeys are non-standard in Segwit Context",
-        );
         Ok(frag)
     }
 }
@@ -509,71 +505,83 @@ where
                 top.args.len(),
             ))),
         }?;
-        // Check whether the unwrapped miniscript is valid under the current context
-        Ctx::check_frag_validity(&unwrapped)?;
         for ch in frag_wrap.chars().rev() {
+            // Check whether the wrapper is valid under the current context
+            let ms = Miniscript::from_ast(unwrapped)?;
+            Ctx::check_global_validity(&ms)?;
             match ch {
-                'a' => unwrapped = Terminal::Alt(Arc::new(Miniscript::from_ast(unwrapped)?)),
-                's' => unwrapped = Terminal::Swap(Arc::new(Miniscript::from_ast(unwrapped)?)),
-                'c' => unwrapped = Terminal::Check(Arc::new(Miniscript::from_ast(unwrapped)?)),
-                'd' => unwrapped = Terminal::DupIf(Arc::new(Miniscript::from_ast(unwrapped)?)),
-                'v' => unwrapped = Terminal::Verify(Arc::new(Miniscript::from_ast(unwrapped)?)),
-                'j' => unwrapped = Terminal::NonZero(Arc::new(Miniscript::from_ast(unwrapped)?)),
-                'n' => {
-                    unwrapped = Terminal::ZeroNotEqual(Arc::new(Miniscript::from_ast(unwrapped)?))
-                }
+                'a' => unwrapped = Terminal::Alt(Arc::new(ms)),
+                's' => unwrapped = Terminal::Swap(Arc::new(ms)),
+                'c' => unwrapped = Terminal::Check(Arc::new(ms)),
+                'd' => unwrapped = Terminal::DupIf(Arc::new(ms)),
+                'v' => unwrapped = Terminal::Verify(Arc::new(ms)),
+                'j' => unwrapped = Terminal::NonZero(Arc::new(ms)),
+                'n' => unwrapped = Terminal::ZeroNotEqual(Arc::new(ms)),
                 't' => {
                     unwrapped = Terminal::AndV(
-                        Arc::new(Miniscript::from_ast(unwrapped)?),
+                        Arc::new(ms),
                         Arc::new(Miniscript::from_ast(Terminal::True)?),
                     )
                 }
                 'u' => {
                     unwrapped = Terminal::OrI(
-                        Arc::new(Miniscript::from_ast(unwrapped)?),
+                        Arc::new(ms),
                         Arc::new(Miniscript::from_ast(Terminal::False)?),
                     )
                 }
                 'l' => {
-                    if unwrapped == Terminal::False {
+                    if ms.node == Terminal::False {
                         return Err(Error::LikelyFalse);
                     }
                     unwrapped = Terminal::OrI(
                         Arc::new(Miniscript::from_ast(Terminal::False)?),
-                        Arc::new(Miniscript::from_ast(unwrapped)?),
+                        Arc::new(ms),
                     )
                 }
                 x => return Err(Error::UnknownWrapper(x)),
             }
-            // Check whether the wrapper is valid under the current context
-            Ctx::check_frag_validity(&unwrapped)?;
         }
-        Ok(unwrapped)
+        // Check whether the unwrapped miniscript is valid under the current context
+        let ms = Miniscript::from_ast(unwrapped)?;
+        Ctx::check_global_validity(&ms)?;
+        Ok(ms.node)
     }
 }
 
 /// Helper trait to add a `push_astelem` method to `script::Builder`
 trait PushAstElem<Pk: MiniscriptKey, Ctx: ScriptContext> {
-    fn push_astelem(self, ast: &Miniscript<Pk, Ctx>) -> Self;
+    fn push_astelem<ToPkCtx: Copy>(self, ast: &Miniscript<Pk, Ctx>, to_pk_ctx: ToPkCtx) -> Self
+    where
+        Pk: ToPublicKey<ToPkCtx>;
 }
 
-impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> PushAstElem<Pk, Ctx> for script::Builder {
-    fn push_astelem(self, ast: &Miniscript<Pk, Ctx>) -> Self {
-        ast.node.encode(self)
+impl<Pk: MiniscriptKey, Ctx: ScriptContext> PushAstElem<Pk, Ctx> for script::Builder {
+    fn push_astelem<ToPkCtx: Copy>(self, ast: &Miniscript<Pk, Ctx>, to_pk_ctx: ToPkCtx) -> Self
+    where
+        Pk: ToPublicKey<ToPkCtx>,
+    {
+        ast.node.encode(self, to_pk_ctx)
     }
 }
 
-impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Terminal<Pk, Ctx> {
+impl<Pk: MiniscriptKey, Ctx: ScriptContext> Terminal<Pk, Ctx> {
     /// Encode the element as a fragment of Bitcoin Script. The inverse
     /// function, from Script to an AST element, is implemented in the
     /// `parse` module.
-    pub fn encode(&self, mut builder: script::Builder) -> script::Builder {
+    pub fn encode<ToPkCtx: Copy>(
+        &self,
+        mut builder: script::Builder,
+        to_pk_ctx: ToPkCtx,
+    ) -> script::Builder
+    where
+        Pk: ToPublicKey<ToPkCtx>,
+    {
         match *self {
-            Terminal::PkK(ref pk) => builder.push_key(&pk.to_public_key()),
+            Terminal::PkK(ref pk) => builder.push_key(&pk.to_public_key(to_pk_ctx)),
             Terminal::PkH(ref hash) => builder
                 .push_opcode(opcodes::all::OP_DUP)
                 .push_opcode(opcodes::all::OP_HASH160)
-                .push_slice(&Pk::hash_to_hash160(&hash)[..])
+                .push_slice(&Pk::hash_to_hash160(&hash, to_pk_ctx)[..])
                 .push_opcode(opcodes::all::OP_EQUALVERIFY),
             Terminal::After(t) => builder
                 .push_int(t as i64)
@@ -611,64 +619,70 @@ impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Terminal<Pk, Ctx> {
             Terminal::False => builder.push_opcode(opcodes::OP_FALSE),
             Terminal::Alt(ref sub) => builder
                 .push_opcode(opcodes::all::OP_TOALTSTACK)
-                .push_astelem(sub)
+                .push_astelem(sub, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_FROMALTSTACK),
-            Terminal::Swap(ref sub) => builder.push_opcode(opcodes::all::OP_SWAP).push_astelem(sub),
+            Terminal::Swap(ref sub) => builder
+                .push_opcode(opcodes::all::OP_SWAP)
+                .push_astelem(sub, to_pk_ctx),
             Terminal::Check(ref sub) => builder
-                .push_astelem(sub)
+                .push_astelem(sub, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_CHECKSIG),
             Terminal::DupIf(ref sub) => builder
                 .push_opcode(opcodes::all::OP_DUP)
                 .push_opcode(opcodes::all::OP_IF)
-                .push_astelem(sub)
+                .push_astelem(sub, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_ENDIF),
-            Terminal::Verify(ref sub) => builder.push_astelem(sub).push_verify(),
+            Terminal::Verify(ref sub) => builder.push_astelem(sub, to_pk_ctx).push_verify(),
             Terminal::NonZero(ref sub) => builder
                 .push_opcode(opcodes::all::OP_SIZE)
                 .push_opcode(opcodes::all::OP_0NOTEQUAL)
                 .push_opcode(opcodes::all::OP_IF)
-                .push_astelem(sub)
+                .push_astelem(sub, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_ENDIF),
             Terminal::ZeroNotEqual(ref sub) => builder
-                .push_astelem(sub)
+                .push_astelem(sub, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_0NOTEQUAL),
-            Terminal::AndV(ref left, ref right) => builder.push_astelem(left).push_astelem(right),
+            Terminal::AndV(ref left, ref right) => builder
+                .push_astelem(left, to_pk_ctx)
+                .push_astelem(right, to_pk_ctx),
             Terminal::AndB(ref left, ref right) => builder
-                .push_astelem(left)
-                .push_astelem(right)
+                .push_astelem(left, to_pk_ctx)
+                .push_astelem(right, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_BOOLAND),
             Terminal::AndOr(ref a, ref b, ref c) => builder
-                .push_astelem(a)
+                .push_astelem(a, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_NOTIF)
-                .push_astelem(c)
+                .push_astelem(c, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_ELSE)
-                .push_astelem(b)
+                .push_astelem(b, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_ENDIF),
             Terminal::OrB(ref left, ref right) => builder
-                .push_astelem(left)
-                .push_astelem(right)
+                .push_astelem(left, to_pk_ctx)
+                .push_astelem(right, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_BOOLOR),
             Terminal::OrD(ref left, ref right) => builder
-                .push_astelem(left)
+                .push_astelem(left, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_IFDUP)
                 .push_opcode(opcodes::all::OP_NOTIF)
-                .push_astelem(right)
+                .push_astelem(right, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_ENDIF),
             Terminal::OrC(ref left, ref right) => builder
-                .push_astelem(left)
+                .push_astelem(left, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_NOTIF)
-                .push_astelem(right)
+                .push_astelem(right, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_ENDIF),
             Terminal::OrI(ref left, ref right) => builder
                 .push_opcode(opcodes::all::OP_IF)
-                .push_astelem(left)
+                .push_astelem(left, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_ELSE)
-                .push_astelem(right)
+                .push_astelem(right, to_pk_ctx)
                 .push_opcode(opcodes::all::OP_ENDIF),
             Terminal::Thresh(k, ref subs) => {
-                builder = builder.push_astelem(&subs[0]);
+                builder = builder.push_astelem(&subs[0], to_pk_ctx);
                 for sub in &subs[1..] {
-                    builder = builder.push_astelem(sub).push_opcode(opcodes::all::OP_ADD);
+                    builder = builder
+                        .push_astelem(sub, to_pk_ctx)
+                        .push_opcode(opcodes::all::OP_ADD);
                 }
                 builder
                     .push_int(k as i64)
@@ -677,7 +691,7 @@ impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Terminal<Pk, Ctx> {
             Terminal::Multi(k, ref keys) => {
                 builder = builder.push_int(k as i64);
                 for pk in keys {
-                    builder = builder.push_key(&pk.to_public_key());
+                    builder = builder.push_key(&pk.to_public_key(to_pk_ctx));
                 }
                 builder
                     .push_int(keys.len() as i64)
@@ -693,9 +707,12 @@ impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Terminal<Pk, Ctx> {
     /// In general, it is not recommended to use this function directly, but
     /// to instead call the corresponding function on a `Descriptor`, which
     /// will handle the segwit/non-segwit technicalities for you.
-    pub fn script_size(&self) -> usize {
+    pub fn script_size<ToPkCtx: Copy>(&self, to_pk_ctx: ToPkCtx) -> usize
+    where
+        Pk: ToPublicKey<ToPkCtx>,
+    {
         match *self {
-            Terminal::PkK(ref pk) => pk.serialized_len(),
+            Terminal::PkK(ref pk) => pk.serialized_len(to_pk_ctx),
             Terminal::PkH(..) => 24,
             Terminal::After(n) => script_num_size(n as usize) + 1,
             Terminal::Older(n) => script_num_size(n as usize) + 1,
@@ -705,29 +722,44 @@ impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Terminal<Pk, Ctx> {
             Terminal::Hash160(..) => 21 + 6,
             Terminal::True => 1,
             Terminal::False => 1,
-            Terminal::Alt(ref sub) => sub.node.script_size() + 2,
-            Terminal::Swap(ref sub) => sub.node.script_size() + 1,
-            Terminal::Check(ref sub) => sub.node.script_size() + 1,
-            Terminal::DupIf(ref sub) => sub.node.script_size() + 3,
+            Terminal::Alt(ref sub) => sub.node.script_size(to_pk_ctx) + 2,
+            Terminal::Swap(ref sub) => sub.node.script_size(to_pk_ctx) + 1,
+            Terminal::Check(ref sub) => sub.node.script_size(to_pk_ctx) + 1,
+            Terminal::DupIf(ref sub) => sub.node.script_size(to_pk_ctx) + 3,
             Terminal::Verify(ref sub) => {
-                sub.node.script_size() + if sub.ext.has_free_verify { 0 } else { 1 }
+                sub.node.script_size(to_pk_ctx) + if sub.ext.has_free_verify { 0 } else { 1 }
+            }
+            Terminal::NonZero(ref sub) => sub.node.script_size(to_pk_ctx) + 4,
+            Terminal::ZeroNotEqual(ref sub) => sub.node.script_size(to_pk_ctx) + 1,
+            Terminal::AndV(ref l, ref r) => {
+                l.node.script_size(to_pk_ctx) + r.node.script_size(to_pk_ctx)
+            }
+            Terminal::AndB(ref l, ref r) => {
+                l.node.script_size(to_pk_ctx) + r.node.script_size(to_pk_ctx) + 1
             }
-            Terminal::NonZero(ref sub) => sub.node.script_size() + 4,
-            Terminal::ZeroNotEqual(ref sub) => sub.node.script_size() + 1,
-            Terminal::AndV(ref l, ref r) => l.node.script_size() + r.node.script_size(),
-            Terminal::AndB(ref l, ref r) => l.node.script_size() + r.node.script_size() + 1,
             Terminal::AndOr(ref a, ref b, ref c) => {
-                a.node.script_size() + b.node.script_size() + c.node.script_size() + 3
+                a.node.script_size(to_pk_ctx)
+                    + b.node.script_size(to_pk_ctx)
+                    + c.node.script_size(to_pk_ctx)
+                    + 3
+            }
+            Terminal::OrB(ref l, ref r) => {
+                l.node.script_size(to_pk_ctx) + r.node.script_size(to_pk_ctx) + 1
+            }
+            Terminal::OrD(ref l, ref r) => {
+                l.node.script_size(to_pk_ctx) + r.node.script_size(to_pk_ctx) + 3
+            }
+            Terminal::OrC(ref l, ref r) => {
+                l.node.script_size(to_pk_ctx) + r.node.script_size(to_pk_ctx) + 2
+            }
+            Terminal::OrI(ref l, ref r) => {
+                l.node.script_size(to_pk_ctx) + r.node.script_size(to_pk_ctx) + 3
             }
-            Terminal::OrB(ref l, ref r) => l.node.script_size() + r.node.script_size() + 1,
-            Terminal::OrD(ref l, ref r) => l.node.script_size() + r.node.script_size() + 3,
-            Terminal::OrC(ref l, ref r) => l.node.script_size() + r.node.script_size() + 2,
-            Terminal::OrI(ref l, ref r) => l.node.script_size() + r.node.script_size() + 3,
             Terminal::Thresh(k, ref subs) => {
                 assert!(!subs.is_empty(), "threshold must be nonempty");
                 script_num_size(k) // k
                     + 1 // EQUAL
-                    + subs.iter().map(|s| s.node.script_size()).sum::<usize>()
+                    + subs.iter().map(|s| s.node.script_size(to_pk_ctx)).sum::<usize>()
                     + subs.len() // ADD
                     - 1 // no ADD on first element
             }
@@ -735,260 +767,11 @@ impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Terminal<Pk, Ctx> {
                 script_num_size(k)
                     + 1
                     + script_num_size(pks.len())
-                    + pks.iter().map(ToPublicKey::serialized_len).sum::<usize>()
+                    + pks
+                        .iter()
+                        .map(|pk| pk.serialized_len(to_pk_ctx))
+                        .sum::<usize>()
             }
         }
     }
-
-    /// Maximum number of witness elements used to dissatisfy the Miniscript
-    /// fragment. Used to estimate the weight of the `VarInt` that specifies
-    /// this number in a serialized transaction.
-    ///
-    /// Will panic if the fragment is not an E, W or Ke.
-    pub fn max_dissatisfaction_witness_elements(&self) -> Option<usize> {
-        match *self {
-            Terminal::PkK(..) => Some(1),
-            Terminal::PkH(..) => Some(2),
-            Terminal::False => Some(0),
-            Terminal::Alt(ref sub) | Terminal::Swap(ref sub) | Terminal::Check(ref sub) => {
-                sub.node.max_dissatisfaction_witness_elements()
-            }
-            Terminal::DupIf(..) | Terminal::NonZero(..) => Some(1),
-            Terminal::AndB(ref l, ref r) => Some(
-                l.node.max_dissatisfaction_witness_elements()?
-                    + r.node.max_dissatisfaction_witness_elements()?,
-            ),
-            Terminal::AndOr(ref a, _, ref c) => Some(
-                a.node.max_dissatisfaction_witness_elements()?
-                    + c.node.max_dissatisfaction_witness_elements()?,
-            ),
-            Terminal::OrB(ref l, ref r) | Terminal::OrD(ref l, ref r) => Some(
-                l.node.max_dissatisfaction_witness_elements()?
-                    + r.node.max_dissatisfaction_witness_elements()?,
-            ),
-            Terminal::OrI(ref l, ref r) => match (
-                l.node.max_dissatisfaction_witness_elements(),
-                r.node.max_dissatisfaction_witness_elements(),
-            ) {
-                (None, Some(r)) => Some(1 + r),
-                (Some(l), None) => Some(1 + l),
-                (None, None) => None,
-                (..) => panic!("tried to dissatisfy or_i with both branches being dissatisfiable"),
-            },
-            Terminal::Thresh(_, ref subs) => {
-                let mut sum = 0;
-                for sub in subs {
-                    match sub.node.max_dissatisfaction_witness_elements() {
-                        Some(s) => sum += s,
-                        None => return None,
-                    }
-                }
-                Some(sum)
-            }
-            Terminal::Multi(k, _) => Some(1 + k),
-            _ => None,
-        }
-    }
-
-    /// Maximum dissatisfaction cost, in bytes, of a Miniscript fragment,
-    /// if it is possible to compute this. This function should probably
-    /// not ever be used directly. It is called from `max_satisfaction_size`.
-    ///
-    /// Will panic if the fragment is not E, W or Ke
-    pub fn max_dissatisfaction_size(&self, one_cost: usize) -> Option<usize> {
-        match *self {
-            Terminal::PkK(..) => Some(1),
-            Terminal::PkH(..) => Some(35),
-            Terminal::False => Some(0),
-            Terminal::Alt(ref sub) | Terminal::Swap(ref sub) | Terminal::Check(ref sub) => {
-                sub.node.max_dissatisfaction_size(one_cost)
-            }
-            Terminal::DupIf(..) | Terminal::NonZero(..) => Some(1),
-            Terminal::AndB(ref l, ref r) => Some(
-                l.node.max_dissatisfaction_size(one_cost)?
-                    + r.node.max_dissatisfaction_size(one_cost)?,
-            ),
-            Terminal::AndOr(ref a, _, ref c) => Some(
-                a.node.max_dissatisfaction_size(one_cost)?
-                    + c.node.max_dissatisfaction_size(one_cost)?,
-            ),
-            Terminal::OrB(ref l, ref r) | Terminal::OrD(ref l, ref r) => Some(
-                l.node.max_dissatisfaction_size(one_cost)?
-                    + r.node.max_dissatisfaction_size(one_cost)?,
-            ),
-            Terminal::OrI(ref l, ref r) => match (
-                l.node.max_dissatisfaction_size(one_cost),
-                r.node.max_dissatisfaction_size(one_cost),
-            ) {
-                (None, Some(r)) => Some(1 + r),
-                (Some(l), None) => Some(one_cost + l),
-                (None, None) => None,
-                (..) => panic!("tried to dissatisfy or_i with both branches being dissatisfiable"),
-            },
-            Terminal::Thresh(_, ref subs) => {
-                let mut sum = 0;
-                for sub in subs {
-                    match sub.node.max_dissatisfaction_size(one_cost) {
-                        Some(s) => sum += s,
-                        None => return None,
-                    }
-                }
-                Some(sum)
-            }
-            Terminal::Multi(k, _) => Some(1 + k),
-            _ => None,
-        }
-    }
-
-    /// Maximum number of witness elements used to satisfy the Miniscript
-    /// fragment. Used to estimate the weight of the `VarInt` that specifies
-    /// this number in a serialized transaction.
-    ///
-    /// This number does not include the witness script itself, so 1 needs
-    /// to be added to the final result.
-    pub fn max_satisfaction_witness_elements(&self) -> usize {
-        match *self {
-            Terminal::PkK(..) => 1,
-            Terminal::PkH(..) => 2,
-            Terminal::After(..) | Terminal::Older(..) => 0,
-            Terminal::Sha256(..)
-            | Terminal::Hash256(..)
-            | Terminal::Ripemd160(..)
-            | Terminal::Hash160(..) => 1,
-            Terminal::True => 0,
-            Terminal::False => 0,
-            Terminal::Alt(ref sub) | Terminal::Swap(ref sub) | Terminal::Check(ref sub) => {
-                sub.node.max_satisfaction_witness_elements()
-            }
-            Terminal::DupIf(ref sub) => 1 + sub.node.max_satisfaction_witness_elements(),
-            Terminal::Verify(ref sub)
-            | Terminal::NonZero(ref sub)
-            | Terminal::ZeroNotEqual(ref sub) => sub.node.max_satisfaction_witness_elements(),
-            Terminal::AndV(ref l, ref r) | Terminal::AndB(ref l, ref r) => {
-                l.node.max_satisfaction_witness_elements()
-                    + r.node.max_satisfaction_witness_elements()
-            }
-            Terminal::AndOr(ref a, ref b, ref c) => cmp::max(
-                a.node.max_satisfaction_witness_elements()
-                    + c.node.max_satisfaction_witness_elements(),
-                a.node.max_dissatisfaction_witness_elements().unwrap()
-                    + b.node.max_satisfaction_witness_elements(),
-            ),
-            Terminal::OrB(ref l, ref r) => cmp::max(
-                l.node.max_satisfaction_witness_elements()
-                    + r.node.max_dissatisfaction_witness_elements().unwrap(),
-                l.node.max_dissatisfaction_witness_elements().unwrap()
-                    + r.node.max_satisfaction_witness_elements(),
-            ),
-            Terminal::OrD(ref l, ref r) | Terminal::OrC(ref l, ref r) => cmp::max(
-                l.node.max_satisfaction_witness_elements(),
-                l.node.max_dissatisfaction_witness_elements().unwrap()
-                    + r.node.max_satisfaction_witness_elements(),
-            ),
-            Terminal::OrI(ref l, ref r) => {
-                1 + cmp::max(
-                    l.node.max_satisfaction_witness_elements(),
-                    r.node.max_satisfaction_witness_elements(),
-                )
-            }
-            Terminal::Thresh(k, ref subs) => {
-                let mut sub_n = subs
-                    .iter()
-                    .map(|sub| {
-                        (
-                            sub.node.max_satisfaction_witness_elements(),
-                            sub.node.max_dissatisfaction_witness_elements().unwrap(),
-                        )
-                    })
-                    .collect::<Vec<(usize, usize)>>();
-                sub_n.sort_by_key(|&(x, y)| x - y);
-                sub_n
-                    .iter()
-                    .rev()
-                    .enumerate()
-                    .map(|(n, &(x, y))| if n < k { x } else { y })
-                    .sum::<usize>()
-            }
-            Terminal::Multi(k, _) => 1 + k,
-        }
-    }
-
-    /// Maximum size, in bytes, of a satisfying witness. For Segwit outputs
-    /// `one_cost` should be set to 2, since the number `1` requires two
-    /// bytes to encode. For non-segwit outputs `one_cost` should be set to
-    /// 1, since `OP_1` is available in scriptSigs.
-    ///
-    /// In general, it is not recommended to use this function directly, but
-    /// to instead call the corresponding function on a `Descriptor`, which
-    /// will handle the segwit/non-segwit technicalities for you.
-    ///
-    /// All signatures are assumed to be 73 bytes in size, including the
-    /// length prefix (segwit) or push opcode (pre-segwit) and sighash
-    /// postfix.
-    ///
-    /// This function may panic on misformed `Miniscript` objects which do not
-    /// correspond to semantically sane Scripts. (Such scripts should be rejected
-    /// at parse time. Any exceptions are bugs.)
-    pub fn max_satisfaction_size(&self, one_cost: usize) -> usize {
-        match *self {
-            Terminal::PkK(..) => 73,
-            Terminal::PkH(..) => 34 + 73,
-            Terminal::After(..) | Terminal::Older(..) => 0,
-            Terminal::Sha256(..)
-            | Terminal::Hash256(..)
-            | Terminal::Ripemd160(..)
-            | Terminal::Hash160(..) => 33,
-            Terminal::True => 0,
-            Terminal::False => 0,
-            Terminal::Alt(ref sub) | Terminal::Swap(ref sub) | Terminal::Check(ref sub) => {
-                sub.node.max_satisfaction_size(one_cost)
-            }
-            Terminal::DupIf(ref sub) => one_cost + sub.node.max_satisfaction_size(one_cost),
-            Terminal::Verify(ref sub)
-            | Terminal::NonZero(ref sub)
-            | Terminal::ZeroNotEqual(ref sub) => sub.node.max_satisfaction_size(one_cost),
-            Terminal::AndV(ref l, ref r) | Terminal::AndB(ref l, ref r) => {
-                l.node.max_satisfaction_size(one_cost) + r.node.max_satisfaction_size(one_cost)
-            }
-            Terminal::AndOr(ref a, ref b, ref c) => cmp::max(
-                a.node.max_satisfaction_size(one_cost) + c.node.max_satisfaction_size(one_cost),
-                a.node.max_dissatisfaction_size(one_cost).unwrap()
-                    + b.node.max_satisfaction_size(one_cost),
-            ),
-            Terminal::OrB(ref l, ref r) => cmp::max(
-                l.node.max_satisfaction_size(one_cost)
-                    + r.node.max_dissatisfaction_size(one_cost).unwrap(),
-                l.node.max_dissatisfaction_size(one_cost).unwrap()
-                    + r.node.max_satisfaction_size(one_cost),
-            ),
-            Terminal::OrD(ref l, ref r) | Terminal::OrC(ref l, ref r) => cmp::max(
-                l.node.max_satisfaction_size(one_cost),
-                l.node.max_dissatisfaction_size(one_cost).unwrap()
-                    + r.node.max_satisfaction_size(one_cost),
-            ),
-            Terminal::OrI(ref l, ref r) => cmp::max(
-                one_cost + l.node.max_satisfaction_size(one_cost),
-                1 + r.node.max_satisfaction_size(one_cost),
-            ),
-            Terminal::Thresh(k, ref subs) => {
-                let mut sub_n = subs
-                    .iter()
-                    .map(|sub| {
-                        (
-                            sub.node.max_satisfaction_size(one_cost),
-                            sub.node.max_dissatisfaction_size(one_cost).unwrap(),
-                        )
-                    })
-                    .collect::<Vec<(usize, usize)>>();
-                sub_n.sort_by_key(|&(x, y)| x - y);
-                sub_n
-                    .iter()
-                    .rev()
-                    .enumerate()
-                    .map(|(n, &(x, y))| if n < k { x } else { y })
-                    .sum::<usize>()
-            }
-            Terminal::Multi(k, _) => 1 + 73 * k,
-        }
-    }
 }
