@@ -30,14 +30,15 @@ use std::{fmt, str};
 use bitcoin;
 use bitcoin::blockdata::script;
 
-pub use self::context::Legacy;
-pub use self::context::Segwitv0;
+pub use self::context::{Bare, Legacy, Segwitv0};
 
+pub mod analyzable;
 pub mod astelem;
 pub(crate) mod context;
 pub mod decode;
 pub mod iter;
 pub mod lex;
+pub mod limits;
 pub mod satisfy;
 pub mod types;
 
@@ -130,23 +131,28 @@ impl<Pk: MiniscriptKey, Ctx: ScriptContext> Miniscript<Pk, Ctx> {
         self.node
     }
 
+    /// Get a reference to the inner `AstElem` representing the root of miniscript
     pub fn as_inner(&self) -> &Terminal<Pk, Ctx> {
         &self.node
     }
 }
 
 impl<Ctx: ScriptContext> Miniscript<bitcoin::PublicKey, Ctx> {
-    /// Attempt to parse a script into a Miniscript representation
-    pub fn parse(script: &script::Script) -> Result<Miniscript<bitcoin::PublicKey, Ctx>, Error> {
-        // Transactions more than 100Kb are non-standard
-        if script.len() > 100_000 {
-            return Err(Error::ScriptSizeTooLarge);
-        }
+    /// Attempt to parse an insane(scripts don't clear sanity checks)
+    /// script into a Miniscript representation.
+    /// Use this to parse scripts with repeated pubkeys, timelock mixing, malleable
+    /// scripts without sig or scripts that can exceed resource limits.
+    /// Some of the analysis guarantees of miniscript are lost when dealing with
+    /// insane scripts. In general, in a multi-party setting users should only
+    /// accept sane scripts.
+    pub fn parse_insane(
+        script: &script::Script,
+    ) -> Result<Miniscript<bitcoin::PublicKey, Ctx>, Error> {
         let tokens = lex(script)?;
         let mut iter = TokenIter::new(tokens);
 
         let top = decode::parse(&mut iter)?;
-        Ctx::check_frag_validity(&top.node)?;
+        Ctx::check_global_validity(&top)?;
         let type_check = types::Type::type_check(&top.node, |_| None)?;
         if type_check.corr.base != types::Base::B {
             return Err(Error::NonTopLevel(format!("{:?}", top)));
@@ -157,12 +163,31 @@ impl<Ctx: ScriptContext> Miniscript<bitcoin::PublicKey, Ctx> {
             Ok(top)
         }
     }
+
+    /// Attempt to parse a Script into Miniscript representation.
+    /// This function will fail parsing for scripts that do not clear
+    /// the [fn.analyzable.sanity_check] checks. Use [fn.parse_insane] to
+    /// parse such scripts.
+    pub fn parse(script: &script::Script) -> Result<Miniscript<bitcoin::PublicKey, Ctx>, Error> {
+        let ms = Self::parse_insane(script)?;
+        ms.sanity_check()?;
+        Ok(ms)
+    }
 }
 
-impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Miniscript<Pk, Ctx> {
+impl<Pk, Ctx> Miniscript<Pk, Ctx>
+where
+    Pk: MiniscriptKey,
+    Ctx: ScriptContext,
+{
     /// Encode as a Bitcoin script
-    pub fn encode(&self) -> script::Script {
-        self.node.encode(script::Builder::new()).into_script()
+    pub fn encode<ToPkCtx: Copy>(&self, to_pk_ctx: ToPkCtx) -> script::Script
+    where
+        Pk: ToPublicKey<ToPkCtx>,
+    {
+        self.node
+            .encode(script::Builder::new(), to_pk_ctx)
+            .into_script()
     }
 
     /// Size, in bytes of the script-pubkey. If this Miniscript is used outside
@@ -172,20 +197,25 @@ impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Miniscript<Pk, Ctx> {
     /// In general, it is not recommended to use this function directly, but
     /// to instead call the corresponding function on a `Descriptor`, which
     /// will handle the segwit/non-segwit technicalities for you.
-    pub fn script_size(&self) -> usize {
-        self.node.script_size()
+    pub fn script_size<ToPkCtx: Copy>(&self, to_pk_ctx: ToPkCtx) -> usize
+    where
+        Pk: ToPublicKey<ToPkCtx>,
+    {
+        self.node.script_size(to_pk_ctx)
     }
+}
 
+impl<Pk: MiniscriptKey, Ctx: ScriptContext> Miniscript<Pk, Ctx> {
     /// Maximum number of witness elements used to satisfy the Miniscript
     /// fragment, including the witness script itself. Used to estimate
     /// the weight of the `VarInt` that specifies this number in a serialized
     /// transaction.
     ///
-    /// This function may panic on misformed `Miniscript` objects which do
+    /// This function may return None on malformed `Miniscript` objects which do
     /// not correspond to semantically sane Scripts. (Such scripts should be
     /// rejected at parse time. Any exceptions are bugs.)
-    pub fn max_satisfaction_witness_elements(&self) -> usize {
-        1 + self.node.max_satisfaction_witness_elements()
+    pub fn max_satisfaction_witness_elements(&self) -> Option<usize> {
+        self.ext.stack_elem_count_sat.map(|x| x + 1)
     }
 
     /// Maximum size, in bytes, of a satisfying witness. For Segwit outputs
@@ -200,12 +230,8 @@ impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Miniscript<Pk, Ctx> {
     /// All signatures are assumed to be 73 bytes in size, including the
     /// length prefix (segwit) or push opcode (pre-segwit) and sighash
     /// postfix.
-    ///
-    /// This function may panic on misformed `Miniscript` objects which do not
-    /// correspond to semantically sane Scripts. (Such scripts should be
-    /// rejected at parse time. Any exceptions are bugs.)
-    pub fn max_satisfaction_size(&self, one_cost: usize) -> usize {
-        self.node.max_satisfaction_size(one_cost)
+    pub fn max_satisfaction_size(&self) -> Option<usize> {
+        Ctx::max_satisfaction_size(self)
     }
 }
 
@@ -224,24 +250,95 @@ impl<Pk: MiniscriptKey, Ctx: ScriptContext> Miniscript<Pk, Ctx> {
         Q: MiniscriptKey,
     {
         let inner = self.node.translate_pk(translatefpk, translatefpkh)?;
-        Ok(Miniscript {
+        let ms = Miniscript {
             //directly copying the type and ext is safe because translating public
             //key should not change any properties
             ty: self.ty,
             ext: self.ext,
             node: inner,
             phantom: PhantomData,
-        })
+        };
+        Ok(ms)
+    }
+
+    /// Attempt to parse an insane(scripts don't clear sanity checks)
+    /// from string into a Miniscript representation.
+    /// Use this to parse scripts with repeated pubkeys, timelock mixing, malleable
+    /// scripts without sig or scripts that can exceed resource limits.
+    /// Some of the analysis guarantees of miniscript are lost when dealing with
+    /// insane scripts. In general, in a multi-party setting users should only
+    /// accept sane scripts.
+    pub fn from_str_insane(s: &str) -> Result<Miniscript<Pk, Ctx>, Error>
+    where
+        <Pk as str::FromStr>::Err: ToString,
+        <<Pk as MiniscriptKey>::Hash as str::FromStr>::Err: ToString,
+    {
+        for ch in s.as_bytes() {
+            if *ch < 20 || *ch > 127 {
+                return Err(Error::Unprintable(*ch));
+            }
+        }
+
+        let top = expression::Tree::from_str(s)?;
+        let ms: Miniscript<Pk, Ctx> = expression::FromTree::from_tree(&top)?;
+
+        if ms.ty.corr.base != types::Base::B {
+            Err(Error::NonTopLevel(format!("{:?}", ms)))
+        } else {
+            Ok(ms)
+        }
     }
 }
 
-impl<Pk: MiniscriptKey + ToPublicKey, Ctx: ScriptContext> Miniscript<Pk, Ctx> {
-    /// Attempt to produce a satisfying witness for the
+impl<Pk: MiniscriptKey, Ctx: ScriptContext> Miniscript<Pk, Ctx> {
+    /// Attempt to produce non-malleable satisfying witness for the
     /// witness script represented by the parse tree
-    pub fn satisfy<S: satisfy::Satisfier<Pk>>(&self, satisfier: S) -> Option<Vec<Vec<u8>>> {
-        match satisfy::Satisfaction::satisfy(&self.node, &satisfier).stack {
-            satisfy::Witness::Stack(stack) => Some(stack),
-            satisfy::Witness::Unavailable => None,
+    pub fn satisfy<ToPkCtx: Copy, S: satisfy::Satisfier<ToPkCtx, Pk>>(
+        &self,
+        satisfier: S,
+        to_pk_ctx: ToPkCtx,
+    ) -> Result<Vec<Vec<u8>>, Error>
+    where
+        Pk: ToPublicKey<ToPkCtx>,
+    {
+        match satisfy::Satisfaction::satisfy(&self.node, &satisfier, self.ty.mall.safe, to_pk_ctx)
+            .stack
+        {
+            satisfy::Witness::Stack(stack) => {
+                Ctx::check_witness::<Pk, Ctx>(&stack)?;
+                Ok(stack)
+            }
+            satisfy::Witness::Unavailable | satisfy::Witness::Impossible => {
+                Err(Error::CouldNotSatisfy)
+            }
+        }
+    }
+
+    /// Attempt to produce a malleable satisfying witness for the
+    /// witness script represented by the parse tree
+    pub fn satisfy_malleable<ToPkCtx: Copy, S: satisfy::Satisfier<ToPkCtx, Pk>>(
+        &self,
+        satisfier: S,
+        to_pk_ctx: ToPkCtx,
+    ) -> Result<Vec<Vec<u8>>, Error>
+    where
+        Pk: ToPublicKey<ToPkCtx>,
+    {
+        match satisfy::Satisfaction::satisfy_mall(
+            &self.node,
+            &satisfier,
+            self.ty.mall.safe,
+            to_pk_ctx,
+        )
+        .stack
+        {
+            satisfy::Witness::Stack(stack) => {
+                Ctx::check_witness::<Pk, Ctx>(&stack)?;
+                Ok(stack)
+            }
+            satisfy::Witness::Unavailable | satisfy::Witness::Impossible => {
+                Err(Error::CouldNotSatisfy)
+            }
         }
     }
 }
@@ -287,21 +384,13 @@ where
 {
     type Err = Error;
 
+    /// Parse a Miniscript from string and perform sanity checks
+    /// See [fn.from_str_insane] to parse scripts from string that
+    /// do not clear the [fn.analyzable.sanity_check] checks.
     fn from_str(s: &str) -> Result<Miniscript<Pk, Ctx>, Error> {
-        for ch in s.as_bytes() {
-            if *ch < 20 || *ch > 127 {
-                return Err(Error::Unprintable(*ch));
-            }
-        }
-
-        let top = expression::Tree::from_str(s)?;
-        let ms: Miniscript<Pk, Ctx> = expression::FromTree::from_tree(&top)?;
-
-        if ms.ty.corr.base != types::Base::B {
-            Err(Error::NonTopLevel(format!("{:?}", ms)))
-        } else {
-            Ok(ms)
-        }
+        let ms = Self::from_str_insane(s)?;
+        ms.sanity_check()?;
+        Ok(ms)
     }
 }
 
@@ -325,6 +414,7 @@ mod tests {
     use std::str::FromStr;
     use std::sync::Arc;
     use MiniscriptKey;
+    use NullCtx;
 
     type Segwitv0Script = Miniscript<bitcoin::PublicKey, Segwitv0>;
 
@@ -380,21 +470,22 @@ mod tests {
 
     fn script_rtt<Str1: Into<Option<&'static str>>>(script: Segwitv0Script, expected_hex: Str1) {
         assert_eq!(script.ty.corr.base, types::Base::B);
-        let bitcoin_script = script.encode();
-        assert_eq!(bitcoin_script.len(), script.script_size());
+        let bitcoin_script = script.encode(NullCtx);
+        assert_eq!(bitcoin_script.len(), script.script_size(NullCtx));
         if let Some(expected) = expected_hex.into() {
             assert_eq!(format!("{:x}", bitcoin_script), expected);
         }
-        let roundtrip = Segwitv0Script::parse(&bitcoin_script).expect("parse string serialization");
+        let roundtrip =
+            Segwitv0Script::parse_insane(&bitcoin_script).expect("parse string serialization");
         assert_eq!(roundtrip, script);
     }
 
     fn roundtrip(tree: &Segwitv0Script, s: &str) {
         assert_eq!(tree.ty.corr.base, types::Base::B);
-        let ser = tree.encode();
-        assert_eq!(ser.len(), tree.script_size());
+        let ser = tree.encode(NullCtx);
+        assert_eq!(ser.len(), tree.script_size(NullCtx));
         assert_eq!(ser.to_string(), s);
-        let deser = Segwitv0Script::parse(&ser).expect("deserialize result of serialize");
+        let deser = Segwitv0Script::parse_insane(&ser).expect("deserialize result of serialize");
         assert_eq!(*tree, deser);
     }
 
@@ -407,10 +498,10 @@ mod tests {
         ops: usize,
         _stack: usize,
     ) {
-        let ms: Result<Segwitv0Script, _> = Miniscript::from_str(ms);
+        let ms: Result<Segwitv0Script, _> = Miniscript::from_str_insane(ms);
         match (ms, valid) {
             (Ok(ms), true) => {
-                assert_eq!(format!("{:x}", ms.encode()), expected_hex);
+                assert_eq!(format!("{:x}", ms.encode(NullCtx)), expected_hex);
                 assert_eq!(ms.ty.mall.non_malleable, non_mal);
                 assert_eq!(ms.ty.mall.safe, need_sig);
                 assert_eq!(ms.ext.ops_count_sat.unwrap(), ops);
@@ -454,7 +545,7 @@ mod tests {
         ms_attributes_test("thresh(2,multi(2,03a0434d9e47f3c86235477c7b1ae6ae5d3442d49b1943c2b752a68e2a47e247c7,036d2b085e9e382ed10b69fc311a03f8641ccfff21574de0927513a49d9a688a00),a:multi(1,036d2b085e9e382ed10b69fc311a03f8641ccfff21574de0927513a49d9a688a00),ac:pk_k(022f01e5e15cca351daff3843fb70f3c2f0a1bdd05e5af888a67784ef3e10a2a01))", "522103a0434d9e47f3c86235477c7b1ae6ae5d3442d49b1943c2b752a68e2a47e247c721036d2b085e9e382ed10b69fc311a03f8641ccfff21574de0927513a49d9a688a0052ae6b5121036d2b085e9e382ed10b69fc311a03f8641ccfff21574de0927513a49d9a688a0051ae6c936b21022f01e5e15cca351daff3843fb70f3c2f0a1bdd05e5af888a67784ef3e10a2a01ac6c935287", true, true, true, 13, 6);
         ms_attributes_test("and_n(sha256(d1ec675902ef1633427ca360b290b0b3045a0d9058ddb5e648b4c3c3224c5c68),t:or_i(v:older(4252898),v:older(144)))", "82012088a820d1ec675902ef1633427ca360b290b0b3045a0d9058ddb5e648b4c3c3224c5c68876400676303e2e440b26967029000b269685168", true, false, false, 14, 2);
         ms_attributes_test("or_d(nd:and_v(v:older(4252898),v:older(4252898)),sha256(38df1c1f64a24a77b23393bca50dff872e31edc4f3b5aa3b90ad0b82f4f089b6))", "766303e2e440b26903e2e440b2696892736482012088a82038df1c1f64a24a77b23393bca50dff872e31edc4f3b5aa3b90ad0b82f4f089b68768", true, false, false, 15, 2);
-        ms_attributes_test("c:and_v(or_c(sha256(9267d3dbed802941483f1afa2a6bc68de5f653128aca9bf1461c5d0a3ad36ed2),v:multi(1,02c44d12c7065d812e8acf28d7cbb19f9011ecd9e9fdf281b0e6a3b5e87d22e7db)),pk_k(03acd484e2f0c7f65309ad178a9f559abde09796974c57e714c35f110dfc27ccbe))", "82012088a8209267d3dbed802941483f1afa2a6bc68de5f653128aca9bf1461c5d0a3ad36ed28764512102c44d12c7065d812e8acf28d7cbb19f9011ecd9e9fdf281b0e6a3b5e87d22e7db51af682103acd484e2f0c7f65309ad178a9f559abde09796974c57e714c35f110dfc27ccbeac", true, false, true, 8, 2);
+        ms_attributes_test("c:and_v(or_c(sha256(9267d3dbed802941483f1afa2a6bc68de5f653128aca9bf1461c5d0a3ad36ed2),v:multi(1,02c44d12c7065d812e8acf28d7cbb19f9011ecd9e9fdf281b0e6a3b5e87d22e7db)),pk_k(03acd484e2f0c7f65309ad178a9f559abde09796974c57e714c35f110dfc27ccbe))", "82012088a8209267d3dbed802941483f1afa2a6bc68de5f653128aca9bf1461c5d0a3ad36ed28764512102c44d12c7065d812e8acf28d7cbb19f9011ecd9e9fdf281b0e6a3b5e87d22e7db51af682103acd484e2f0c7f65309ad178a9f559abde09796974c57e714c35f110dfc27ccbeac", true, false, true, 9, 2);
         ms_attributes_test("c:and_v(or_c(multi(2,036d2b085e9e382ed10b69fc311a03f8641ccfff21574de0927513a49d9a688a00,02352bbf4a4cdd12564f93fa332ce333301d9ad40271f8107181340aef25be59d5),v:ripemd160(1b0f3c404d12075c68c938f9f60ebea4f74941a0)),pk_k(03fff97bd5755eeea420453a14355235d382f6472f8568a18b2f057a1460297556))", "5221036d2b085e9e382ed10b69fc311a03f8641ccfff21574de0927513a49d9a688a002102352bbf4a4cdd12564f93fa332ce333301d9ad40271f8107181340aef25be59d552ae6482012088a6141b0f3c404d12075c68c938f9f60ebea4f74941a088682103fff97bd5755eeea420453a14355235d382f6472f8568a18b2f057a1460297556ac", true, true, true, 10, 5);
         ms_attributes_test("and_v(andor(hash256(8a35d9ca92a48eaade6f53a64985e9e2afeb74dcf8acb4c3721e0dc7e4294b25),v:hash256(939894f70e6c3a25da75da0cc2071b4076d9b006563cf635986ada2e93c0d735),v:older(50000)),after(499999999))", "82012088aa208a35d9ca92a48eaade6f53a64985e9e2afeb74dcf8acb4c3721e0dc7e4294b2587640350c300b2696782012088aa20939894f70e6c3a25da75da0cc2071b4076d9b006563cf635986ada2e93c0d735886804ff64cd1db1", true, false, false, 14, 2);
         ms_attributes_test("andor(hash256(5f8d30e655a7ba0d7596bb3ddfb1d2d20390d23b1845000e1e118b3be1b3f040),j:and_v(v:hash160(3a2bff0da9d96868e66abc4427bea4691cf61ccd),older(4194305)),ripemd160(44d90e2d3714c8663b632fcf0f9d5f22192cc4c8))", "82012088aa205f8d30e655a7ba0d7596bb3ddfb1d2d20390d23b1845000e1e118b3be1b3f040876482012088a61444d90e2d3714c8663b632fcf0f9d5f22192cc4c8876782926382012088a9143a2bff0da9d96868e66abc4427bea4691cf61ccd8803010040b26868", true, false, false, 20, 2);
@@ -555,6 +646,25 @@ mod tests {
         assert!(Segwitv0Script::from_str("tv:1()").is_err());
     }
 
+    #[test]
+    fn verify_parse() {
+        let ms = "and_v(v:hash160(20195b5a3d650c17f0f29f91c33f8f6335193d07),or_d(sha256(96de8fc8c256fa1e1556d41af431cace7dca68707c78dd88c3acab8b17164c47),older(16)))";
+        let ms: Segwitv0Script = Miniscript::from_str_insane(ms).unwrap();
+        assert_eq!(ms, Miniscript::parse_insane(&ms.encode(NullCtx)).unwrap());
+
+        let ms = "and_v(v:sha256(96de8fc8c256fa1e1556d41af431cace7dca68707c78dd88c3acab8b17164c47),or_d(sha256(96de8fc8c256fa1e1556d41af431cace7dca68707c78dd88c3acab8b17164c47),older(16)))";
+        let ms: Segwitv0Script = Miniscript::from_str_insane(ms).unwrap();
+        assert_eq!(ms, Miniscript::parse_insane(&ms.encode(NullCtx)).unwrap());
+
+        let ms = "and_v(v:ripemd160(20195b5a3d650c17f0f29f91c33f8f6335193d07),or_d(sha256(96de8fc8c256fa1e1556d41af431cace7dca68707c78dd88c3acab8b17164c47),older(16)))";
+        let ms: Segwitv0Script = Miniscript::from_str_insane(ms).unwrap();
+        assert_eq!(ms, Miniscript::parse_insane(&ms.encode(NullCtx)).unwrap());
+
+        let ms = "and_v(v:hash256(96de8fc8c256fa1e1556d41af431cace7dca68707c78dd88c3acab8b17164c47),or_d(sha256(96de8fc8c256fa1e1556d41af431cace7dca68707c78dd88c3acab8b17164c47),older(16)))";
+        let ms: Segwitv0Script = Miniscript::from_str_insane(ms).unwrap();
+        assert_eq!(ms, Miniscript::parse_insane(&ms.encode(NullCtx)).unwrap());
+    }
+
     #[test]
     fn pk_alias() {
         let pubkey = pubkeys(1)[0];
@@ -745,21 +855,21 @@ mod tests {
     #[test]
     fn deserialize() {
         // Most of these came from fuzzing, hence the increasing lengths
-        assert!(Segwitv0Script::parse(&hex_script("")).is_err()); // empty
-        assert!(Segwitv0Script::parse(&hex_script("00")).is_ok()); // FALSE
-        assert!(Segwitv0Script::parse(&hex_script("51")).is_ok()); // TRUE
-        assert!(Segwitv0Script::parse(&hex_script("69")).is_err()); // VERIFY
-        assert!(Segwitv0Script::parse(&hex_script("0000")).is_err()); //and_v(FALSE,FALSE)
-        assert!(Segwitv0Script::parse(&hex_script("1001")).is_err()); // incomplete push
-        assert!(Segwitv0Script::parse(&hex_script("03990300b2")).is_err()); // non-minimal #
-        assert!(Segwitv0Script::parse(&hex_script("8559b2")).is_err()); // leading bytes
-        assert!(Segwitv0Script::parse(&hex_script("4c0169b2")).is_err()); // non-minimal push
-        assert!(Segwitv0Script::parse(&hex_script("0000af0000ae85")).is_err()); // OR not BOOLOR
+        assert!(Segwitv0Script::parse_insane(&hex_script("")).is_err()); // empty
+        assert!(Segwitv0Script::parse_insane(&hex_script("00")).is_ok()); // FALSE
+        assert!(Segwitv0Script::parse_insane(&hex_script("51")).is_ok()); // TRUE
+        assert!(Segwitv0Script::parse_insane(&hex_script("69")).is_err()); // VERIFY
+        assert!(Segwitv0Script::parse_insane(&hex_script("0000")).is_err()); //and_v(FALSE,FALSE)
+        assert!(Segwitv0Script::parse_insane(&hex_script("1001")).is_err()); // incomplete push
+        assert!(Segwitv0Script::parse_insane(&hex_script("03990300b2")).is_err()); // non-minimal #
+        assert!(Segwitv0Script::parse_insane(&hex_script("8559b2")).is_err()); // leading bytes
+        assert!(Segwitv0Script::parse_insane(&hex_script("4c0169b2")).is_err()); // non-minimal push
+        assert!(Segwitv0Script::parse_insane(&hex_script("0000af0000ae85")).is_err()); // OR not BOOLOR
 
         // misc fuzzer problems
-        assert!(Segwitv0Script::parse(&hex_script("0000000000af")).is_err());
-        assert!(Segwitv0Script::parse(&hex_script("04009a2970af00")).is_err()); // giant CMS key num
-        assert!(Segwitv0Script::parse(&hex_script(
+        assert!(Segwitv0Script::parse_insane(&hex_script("0000000000af")).is_err());
+        assert!(Segwitv0Script::parse_insane(&hex_script("04009a2970af00")).is_err()); // giant CMS key num
+        assert!(Segwitv0Script::parse_insane(&hex_script(
             "2102ffffffffffffffefefefefefefefefefefef394c0fe5b711179e124008584753ac6900"
         ))
         .is_err());
