@@ -1,6 +1,8 @@
 //! Other miscellaneous type properties which are not related to
 //! correctness or malleability.
 
+use miniscript::limits::{HEIGHT_TIME_THRESHOLD, SEQUENCE_LOCKTIME_TYPE_FLAG};
+
 use super::{Error, ErrorKind, Property, ScriptContext};
 use script_num_size;
 use std::cmp;
@@ -8,14 +10,10 @@ use std::iter::once;
 use MiniscriptKey;
 use Terminal;
 
-pub const MAX_OPS_PER_SCRIPT: usize = 201;
-// https://github.com/bitcoin/bitcoin/blob/9ccaee1d5e2e4b79b0a7c29aadb41b97e4741332/src/script/script.h#L39
-pub const HEIGHT_TIME_THRESHOLD: u32 = 500_000_000;
-
 /// Helper struct Whether any satisfaction of this fragment contains any timelocks
 #[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Debug, Hash)]
 pub struct TimeLockInfo {
-    // csv with heights
+    /// csv with heights
     pub csv_with_height: bool,
     /// csv with times
     pub csv_with_time: bool,
@@ -93,10 +91,7 @@ impl TimeLockInfo {
     }
 }
 
-/// Structure representing the extra type properties of a fragment which are
-/// relevant to legacy(pre-segwit) safety and fee estimation. If a fragment is
-/// used in pre-segwit transactions it will only be malleable but still is
-/// correct and sound.
+/// Structure representing the extra type properties of a fragment.
 #[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Debug, Hash)]
 pub struct ExtData {
     /// The number of bytes needed to encode its scriptpubkey
@@ -109,6 +104,19 @@ pub struct ExtData {
     pub ops_count_sat: Option<usize>,
     /// The worst case ops-count for dissatisfying this Miniscript fragment.
     pub ops_count_nsat: Option<usize>,
+    /// The worst case number of stack elements for satisfying this Miniscript fragment.
+    pub stack_elem_count_sat: Option<usize>,
+    /// The worst case number of stack elements for dissatisfying this Miniscript fragment.
+    pub stack_elem_count_dissat: Option<usize>,
+    /// Maximum size, in bytes, of a satisfying witness. First elements is the cost for the
+    /// witness stack, the second one is the cost for scriptSig.
+    /// All signatures are assumed to be 73 bytes in size, including the
+    /// length prefix (segwit) or push opcode (pre-segwit) and sighash
+    /// postfix.
+    pub max_sat_size: Option<(usize, usize)>,
+    /// Maximum dissatisfaction cost, in bytes, of a Miniscript fragment. First elements is
+    /// the cost for the witness stack, the second one is the cost for scriptSig.
+    pub max_dissat_size: Option<(usize, usize)>,
     /// The timelock info about heightlocks and timelocks
     pub timelock_info: TimeLockInfo,
 }
@@ -125,6 +133,10 @@ impl Property for ExtData {
             ops_count_static: 0,
             ops_count_sat: Some(0),
             ops_count_nsat: None,
+            stack_elem_count_sat: Some(0),
+            stack_elem_count_dissat: None,
+            max_sat_size: Some((0, 0)),
+            max_dissat_size: None,
             timelock_info: TimeLockInfo::default(),
         }
     }
@@ -136,6 +148,10 @@ impl Property for ExtData {
             ops_count_static: 0,
             ops_count_sat: None,
             ops_count_nsat: Some(0),
+            stack_elem_count_sat: None,
+            stack_elem_count_dissat: Some(0),
+            max_sat_size: None,
+            max_dissat_size: Some((0, 0)),
             timelock_info: TimeLockInfo::default(),
         }
     }
@@ -147,6 +163,10 @@ impl Property for ExtData {
             ops_count_static: 0,
             ops_count_sat: Some(0),
             ops_count_nsat: Some(0),
+            stack_elem_count_sat: Some(1),
+            stack_elem_count_dissat: Some(1),
+            max_sat_size: Some((73, 73)),
+            max_dissat_size: Some((1, 1)),
             timelock_info: TimeLockInfo::default(),
         }
     }
@@ -158,6 +178,10 @@ impl Property for ExtData {
             ops_count_static: 3,
             ops_count_sat: Some(3),
             ops_count_nsat: Some(3),
+            stack_elem_count_sat: Some(2),
+            stack_elem_count_dissat: Some(2),
+            max_sat_size: Some((34 + 73, 34 + 73)),
+            max_dissat_size: Some((35, 35)),
             timelock_info: TimeLockInfo::default(),
         }
     }
@@ -175,6 +199,10 @@ impl Property for ExtData {
             ops_count_static: 1,
             ops_count_sat: Some(n + 1),
             ops_count_nsat: Some(n + 1),
+            stack_elem_count_sat: Some(k + 1),
+            stack_elem_count_dissat: Some(k + 1),
+            max_sat_size: Some((1 + 73 * k, 1 + 73 * k)),
+            max_dissat_size: Some((1 + k, 1 + k)),
             timelock_info: TimeLockInfo::default(),
         }
     }
@@ -190,7 +218,11 @@ impl Property for ExtData {
             has_free_verify: true,
             ops_count_static: 4,
             ops_count_sat: Some(4),
-            ops_count_nsat: None,
+            ops_count_nsat: Some(4),
+            stack_elem_count_sat: Some(1),
+            stack_elem_count_dissat: Some(1),
+            max_sat_size: Some((33, 33)),
+            max_dissat_size: Some((33, 33)),
             timelock_info: TimeLockInfo::default(),
         }
     }
@@ -201,7 +233,11 @@ impl Property for ExtData {
             has_free_verify: true,
             ops_count_static: 4,
             ops_count_sat: Some(4),
-            ops_count_nsat: None,
+            ops_count_nsat: Some(4),
+            stack_elem_count_sat: Some(1),
+            stack_elem_count_dissat: Some(1),
+            max_sat_size: Some((33, 33)),
+            max_dissat_size: Some((33, 33)),
             timelock_info: TimeLockInfo::default(),
         }
     }
@@ -212,7 +248,11 @@ impl Property for ExtData {
             has_free_verify: true,
             ops_count_static: 4,
             ops_count_sat: Some(4),
-            ops_count_nsat: None,
+            ops_count_nsat: Some(4),
+            stack_elem_count_sat: Some(1),
+            stack_elem_count_dissat: Some(1),
+            max_sat_size: Some((33, 33)),
+            max_dissat_size: Some((33, 33)),
             timelock_info: TimeLockInfo::default(),
         }
     }
@@ -223,7 +263,11 @@ impl Property for ExtData {
             has_free_verify: true,
             ops_count_static: 4,
             ops_count_sat: Some(4),
-            ops_count_nsat: None,
+            ops_count_nsat: Some(4),
+            stack_elem_count_sat: Some(1),
+            stack_elem_count_dissat: Some(1),
+            max_sat_size: Some((33, 33)),
+            max_dissat_size: Some((33, 33)),
             timelock_info: TimeLockInfo::default(),
         }
     }
@@ -239,6 +283,10 @@ impl Property for ExtData {
             ops_count_static: 1,
             ops_count_sat: Some(1),
             ops_count_nsat: None,
+            stack_elem_count_sat: Some(0),
+            stack_elem_count_dissat: None,
+            max_sat_size: Some((0, 0)),
+            max_dissat_size: None,
             timelock_info: TimeLockInfo {
                 csv_with_height: false,
                 csv_with_time: false,
@@ -256,9 +304,13 @@ impl Property for ExtData {
             ops_count_static: 1,
             ops_count_sat: Some(1),
             ops_count_nsat: None,
+            stack_elem_count_sat: Some(0),
+            stack_elem_count_dissat: None,
+            max_sat_size: Some((0, 0)),
+            max_dissat_size: None,
             timelock_info: TimeLockInfo {
-                csv_with_height: t < HEIGHT_TIME_THRESHOLD,
-                csv_with_time: t >= HEIGHT_TIME_THRESHOLD,
+                csv_with_height: (t & SEQUENCE_LOCKTIME_TYPE_FLAG) == 0,
+                csv_with_time: (t & SEQUENCE_LOCKTIME_TYPE_FLAG) != 0,
                 cltv_with_height: false,
                 cltv_with_time: false,
                 contains_combination: false,
@@ -273,6 +325,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static + 2,
             ops_count_sat: self.ops_count_sat.map(|x| x + 2),
             ops_count_nsat: self.ops_count_nsat.map(|x| x + 2),
+            stack_elem_count_sat: self.stack_elem_count_sat,
+            stack_elem_count_dissat: self.stack_elem_count_dissat,
+            max_sat_size: self.max_sat_size,
+            max_dissat_size: self.max_dissat_size,
             timelock_info: self.timelock_info,
         })
     }
@@ -284,6 +340,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static + 1,
             ops_count_sat: self.ops_count_sat.map(|x| x + 1),
             ops_count_nsat: self.ops_count_nsat.map(|x| x + 1),
+            stack_elem_count_sat: self.stack_elem_count_sat,
+            stack_elem_count_dissat: self.stack_elem_count_dissat,
+            max_sat_size: self.max_sat_size,
+            max_dissat_size: self.max_dissat_size,
             timelock_info: self.timelock_info,
         })
     }
@@ -295,6 +355,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static + 1,
             ops_count_sat: self.ops_count_sat.map(|x| x + 1),
             ops_count_nsat: self.ops_count_nsat.map(|x| x + 1),
+            stack_elem_count_sat: self.stack_elem_count_sat,
+            stack_elem_count_dissat: self.stack_elem_count_dissat,
+            max_sat_size: self.max_sat_size,
+            max_dissat_size: self.max_dissat_size,
             timelock_info: self.timelock_info,
         })
     }
@@ -306,6 +370,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static + 3,
             ops_count_sat: self.ops_count_sat.map(|x| x + 3),
             ops_count_nsat: Some(self.ops_count_static + 3),
+            stack_elem_count_sat: self.stack_elem_count_sat.map(|x| x + 1),
+            stack_elem_count_dissat: Some(1),
+            max_sat_size: self.max_sat_size.map(|(w, s)| (w + 2, s + 1)),
+            max_dissat_size: Some((1, 1)),
             timelock_info: self.timelock_info,
         })
     }
@@ -318,6 +386,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static + verify_cost,
             ops_count_sat: self.ops_count_sat.map(|x| x + verify_cost),
             ops_count_nsat: None,
+            stack_elem_count_sat: self.stack_elem_count_sat,
+            stack_elem_count_dissat: None,
+            max_sat_size: self.max_sat_size,
+            max_dissat_size: None,
             timelock_info: self.timelock_info,
         })
     }
@@ -329,6 +401,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static + 4,
             ops_count_sat: self.ops_count_sat.map(|x| x + 4),
             ops_count_nsat: Some(self.ops_count_static + 4),
+            stack_elem_count_sat: self.stack_elem_count_sat,
+            stack_elem_count_dissat: Some(1),
+            max_sat_size: self.max_sat_size,
+            max_dissat_size: Some((1, 1)),
             timelock_info: self.timelock_info,
         })
     }
@@ -340,6 +416,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static + 1,
             ops_count_sat: self.ops_count_sat.map(|x| x + 1),
             ops_count_nsat: self.ops_count_nsat.map(|x| x + 1),
+            stack_elem_count_sat: self.stack_elem_count_sat,
+            stack_elem_count_dissat: self.stack_elem_count_dissat,
+            max_sat_size: self.max_sat_size,
+            max_dissat_size: self.max_dissat_size,
             timelock_info: self.timelock_info,
         })
     }
@@ -351,6 +431,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static,
             ops_count_sat: self.ops_count_sat,
             ops_count_nsat: None,
+            stack_elem_count_sat: self.stack_elem_count_sat,
+            stack_elem_count_dissat: None,
+            max_sat_size: self.max_sat_size,
+            max_dissat_size: None,
             timelock_info: self.timelock_info,
         })
     }
@@ -367,6 +451,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static + 3,
             ops_count_sat: self.ops_count_sat.map(|x| x + 3),
             ops_count_nsat: Some(self.ops_count_static + 3),
+            stack_elem_count_sat: self.stack_elem_count_sat.map(|x| x + 1),
+            stack_elem_count_dissat: self.stack_elem_count_dissat.map(|x| x + 1),
+            max_sat_size: self.max_sat_size.map(|(w, s)| (w + 2, s + 1)),
+            max_dissat_size: self.max_dissat_size.map(|(w, s)| (w + 1, s + 1)),
             timelock_info: self.timelock_info,
         })
     }
@@ -378,6 +466,10 @@ impl Property for ExtData {
             ops_count_static: self.ops_count_static + 3,
             ops_count_sat: self.ops_count_sat.map(|x| x + 3),
             ops_count_nsat: Some(self.ops_count_static + 3),
+            stack_elem_count_sat: self.stack_elem_count_sat.map(|x| x + 1),
+            stack_elem_count_dissat: self.stack_elem_count_dissat.map(|x| x + 1),
+            max_sat_size: self.max_sat_size.map(|(w, s)| (w + 1, s + 1)),
+            max_dissat_size: self.max_dissat_size.map(|(w, s)| (w + 2, s + 1)),
             timelock_info: self.timelock_info,
         })
     }
@@ -393,6 +485,18 @@ impl Property for ExtData {
             ops_count_nsat: l
                 .ops_count_nsat
                 .and_then(|x| r.ops_count_nsat.map(|y| x + y + 1)),
+            stack_elem_count_sat: l
+                .stack_elem_count_sat
+                .and_then(|l| r.stack_elem_count_sat.map(|r| l + r)),
+            stack_elem_count_dissat: l
+                .stack_elem_count_dissat
+                .and_then(|l| r.stack_elem_count_dissat.map(|r| l + r)),
+            max_sat_size: l
+                .max_sat_size
+                .and_then(|(lw, ls)| r.max_sat_size.map(|(rw, rs)| (lw + rw, ls + rs))),
+            max_dissat_size: l
+                .max_dissat_size
+                .and_then(|(lw, ls)| r.max_dissat_size.map(|(rw, rs)| (lw + rw, ls + rs))),
             timelock_info: TimeLockInfo::comb_and_timelocks(l.timelock_info, r.timelock_info),
         })
     }
@@ -404,6 +508,14 @@ impl Property for ExtData {
             ops_count_static: l.ops_count_static + r.ops_count_static,
             ops_count_sat: l.ops_count_sat.and_then(|x| r.ops_count_sat.map(|y| x + y)),
             ops_count_nsat: None,
+            stack_elem_count_sat: l
+                .stack_elem_count_sat
+                .and_then(|l| r.stack_elem_count_sat.map(|r| l + r)),
+            stack_elem_count_dissat: None,
+            max_sat_size: l
+                .max_sat_size
+                .and_then(|(lw, ls)| r.max_sat_size.map(|(rw, rs)| (lw + rw, ls + rs))),
+            max_dissat_size: None,
             timelock_info: TimeLockInfo::comb_and_timelocks(l.timelock_info, r.timelock_info),
         })
     }
@@ -422,6 +534,24 @@ impl Property for ExtData {
             ops_count_nsat: l
                 .ops_count_nsat
                 .and_then(|x| r.ops_count_nsat.map(|y| x + y + 1)),
+            stack_elem_count_sat: cmp::max(
+                l.stack_elem_count_sat
+                    .and_then(|l| r.stack_elem_count_dissat.map(|r| l + r)),
+                l.stack_elem_count_dissat
+                    .and_then(|l| r.stack_elem_count_sat.map(|r| l + r)),
+            ),
+            stack_elem_count_dissat: l
+                .stack_elem_count_dissat
+                .and_then(|l| r.stack_elem_count_dissat.map(|r| l + r)),
+            max_sat_size: cmp::max(
+                l.max_sat_size
+                    .and_then(|(lw, ls)| r.max_dissat_size.map(|(rw, rs)| (lw + rw, ls + rs))),
+                l.max_dissat_size
+                    .and_then(|(lw, ls)| r.max_sat_size.map(|(rw, rs)| (lw + rw, ls + rs))),
+            ),
+            max_dissat_size: l
+                .max_dissat_size
+                .and_then(|(lw, ls)| r.max_dissat_size.map(|(rw, rs)| (lw + rw, ls + rs))),
             timelock_info: TimeLockInfo::comb_or_timelocks(l.timelock_info, r.timelock_info),
         })
     }
@@ -439,6 +569,22 @@ impl Property for ExtData {
             ops_count_nsat: l
                 .ops_count_nsat
                 .and_then(|x| r.ops_count_nsat.map(|y| x + y + 3)),
+            stack_elem_count_sat: cmp::max(
+                l.stack_elem_count_sat,
+                l.stack_elem_count_dissat
+                    .and_then(|l_dis| r.stack_elem_count_sat.map(|r_sat| r_sat + l_dis)),
+            ),
+            stack_elem_count_dissat: l
+                .stack_elem_count_dissat
+                .and_then(|l_dis| r.stack_elem_count_dissat.map(|r_dis| r_dis + l_dis)),
+            max_sat_size: cmp::max(
+                l.max_sat_size,
+                l.max_dissat_size
+                    .and_then(|(lw, ls)| r.max_sat_size.map(|(rw, rs)| (lw + rw, ls + rs))),
+            ),
+            max_dissat_size: l
+                .max_dissat_size
+                .and_then(|(lw, ls)| r.max_dissat_size.map(|(rw, rs)| (lw + rw, ls + rs))),
             timelock_info: TimeLockInfo::comb_or_timelocks(l.timelock_info, r.timelock_info),
         })
     }
@@ -454,6 +600,18 @@ impl Property for ExtData {
                     .and_then(|x| l.ops_count_nsat.map(|y| y + x + 2)),
             ),
             ops_count_nsat: None,
+            stack_elem_count_sat: cmp::max(
+                l.stack_elem_count_sat,
+                l.stack_elem_count_dissat
+                    .and_then(|l_dis| r.stack_elem_count_sat.map(|r_sat| r_sat + l_dis)),
+            ),
+            stack_elem_count_dissat: None,
+            max_sat_size: cmp::max(
+                l.max_sat_size,
+                l.max_dissat_size
+                    .and_then(|(lw, ls)| r.max_sat_size.map(|(rw, rs)| (lw + rw, ls + rs))),
+            ),
+            max_dissat_size: None,
             timelock_info: TimeLockInfo::comb_or_timelocks(l.timelock_info, r.timelock_info),
         })
     }
@@ -472,6 +630,31 @@ impl Property for ExtData {
                 (_, Some(x)) | (Some(x), _) => Some(x + 3),
                 (None, None) => None,
             },
+            stack_elem_count_sat: match (l.stack_elem_count_sat, r.stack_elem_count_sat) {
+                (Some(l), Some(r)) => Some(1 + cmp::max(l, r)),
+                (Some(l), None) => Some(1 + l),
+                (None, Some(r)) => Some(1 + r),
+                (None, None) => None,
+            },
+            stack_elem_count_dissat: match (l.stack_elem_count_dissat, r.stack_elem_count_dissat) {
+                (Some(l), Some(r)) => Some(1 + cmp::max(l, r)),
+                (Some(l), None) => Some(1 + l),
+                (None, Some(r)) => Some(1 + r),
+                (None, None) => None,
+            },
+            max_sat_size: cmp::max(
+                l.max_sat_size.and_then(|(w, s)| Some((w + 2, s + 1))),
+                r.max_sat_size.and_then(|(w, s)| Some((w + 1, s + 1))),
+            ),
+            max_dissat_size: match (l.max_dissat_size, r.max_dissat_size) {
+                (Some(l), Some(r)) => {
+                    let max = cmp::max(l, r);
+                    Some((1 + max.0, 1 + max.1))
+                }
+                (None, Some(r)) => Some((1 + r.0, 1 + r.1)),
+                (Some(l), None) => Some((2 + l.0, 1 + l.1)),
+                (None, None) => None,
+            },
             timelock_info: TimeLockInfo::comb_or_timelocks(l.timelock_info, r.timelock_info),
         })
     }
@@ -490,6 +673,24 @@ impl Property for ExtData {
             ops_count_nsat: c
                 .ops_count_nsat
                 .and_then(|z| a.ops_count_nsat.map(|x| x + b.ops_count_static + z + 3)),
+            stack_elem_count_sat: cmp::max(
+                a.stack_elem_count_sat
+                    .and_then(|a| b.stack_elem_count_sat.map(|b| b + a)),
+                a.stack_elem_count_dissat
+                    .and_then(|a_dis| c.stack_elem_count_sat.map(|c| c + a_dis)),
+            ),
+            stack_elem_count_dissat: a
+                .stack_elem_count_dissat
+                .and_then(|a_dis| c.stack_elem_count_dissat.map(|c| c + a_dis)),
+            max_sat_size: cmp::max(
+                a.max_sat_size
+                    .and_then(|(wa, sa)| b.max_sat_size.map(|(wb, sb)| (wa + wb, sa + sb))),
+                a.max_dissat_size
+                    .and_then(|(wa, sa)| c.max_sat_size.map(|(wc, sc)| (wa + wc, sa + sc))),
+            ),
+            max_dissat_size: a
+                .max_dissat_size
+                .and_then(|(wa, sa)| c.max_dissat_size.map(|(wc, sc)| (wa + wc, sa + sc))),
             timelock_info: TimeLockInfo::comb_or_timelocks(
                 TimeLockInfo::comb_and_timelocks(a.timelock_info, b.timelock_info),
                 c.timelock_info,
@@ -508,13 +709,35 @@ impl Property for ExtData {
         let mut ops_count_nsat = Some(0);
         let mut ops_count_sat = Some(0);
         let mut sat_count = 0;
-        let mut timelocks = vec![];
+        let mut timelocks = Vec::with_capacity(n);
+        let mut stack_elem_count_sat_vec = Vec::with_capacity(n);
+        let mut stack_elem_count_sat = Some(0);
+        let mut stack_elem_count_dissat = Some(0);
+        let mut max_sat_size_vec = Vec::with_capacity(n);
+        let mut max_sat_size = Some((0, 0));
+        let mut max_dissat_size = Some((0, 0));
+
         for i in 0..n {
             let sub = sub_ck(i)?;
 
             pk_cost += sub.pk_cost;
             ops_count_static += sub.ops_count_static;
             timelocks.push(sub.timelock_info);
+
+            if let Some(n_items) = sub.stack_elem_count_dissat {
+                stack_elem_count_dissat = stack_elem_count_dissat.map(|x| x + n_items);
+                let sub_dissat_size = sub
+                    .max_dissat_size
+                    .expect("dissat_size is None but not stack_elem?");
+                max_dissat_size =
+                    max_dissat_size.map(|(w, s)| (w + sub_dissat_size.0, s + sub_dissat_size.1));
+            } else {
+                // The thresh is dissatifiable iff all sub policies are dissatifiable
+                stack_elem_count_dissat = None;
+            }
+            stack_elem_count_sat_vec.push((sub.stack_elem_count_sat, sub.stack_elem_count_dissat));
+            max_sat_size_vec.push((sub.max_sat_size, sub.max_sat_size));
+
             match (sub.ops_count_sat, sub.ops_count_nsat) {
                 (Some(x), Some(y)) => {
                     ops_count_sat_vec.push(Some(x as i32 - y as i32));
@@ -529,6 +752,38 @@ impl Property for ExtData {
                 _ => {}
             }
         }
+
+        // We sort by [satisfaction cost - dissatisfaction cost] to make a worst-case (the most
+        // costy satisfaction are satisfied, the most costy dissatisfactions are dissatisfied)
+        // sum of the cost by iterating through the sorted vector *backward*.
+        stack_elem_count_sat_vec.sort_by(|a, b| {
+            a.0.map(|x| a.1.map(|y| x as isize - y as isize))
+                .cmp(&b.0.map(|x| b.1.map(|y| x as isize - y as isize)))
+        });
+        for (i, &(x, y)) in stack_elem_count_sat_vec.iter().rev().enumerate() {
+            stack_elem_count_sat = if i <= k {
+                x.and_then(|x| stack_elem_count_sat.map(|count| count + x))
+            } else {
+                y.and_then(|y| stack_elem_count_sat.map(|count| count + y))
+            };
+        }
+
+        // Same for the size cost. A bit more intricated as we need to account for both the witness
+        // and scriptSig cost, so we end up with a tuple of Options of tuples. We use the witness
+        // cost (first element of the mentioned tuple) here.
+        // FIXME: Maybe make the ExtData struct aware of Ctx and add a one_cost() method here ?
+        max_sat_size_vec.sort_by(|a, b| {
+            a.0.map(|x| a.1.map(|y| x.0 as isize - y.0 as isize))
+                .cmp(&b.0.map(|x| b.1.map(|y| x.0 as isize - y.0 as isize)))
+        });
+        for (i, &(x, y)) in max_sat_size_vec.iter().enumerate() {
+            max_sat_size = if i <= k {
+                x.and_then(|x| max_sat_size.map(|(w, s)| (w + x.0, s + x.1)))
+            } else {
+                y.and_then(|y| max_sat_size.map(|(w, s)| (w + y.0, s + y.1)))
+            };
+        }
+
         let remaining_sat = k - sat_count;
         let mut sum: i32 = 0;
         if k < sat_count || ops_count_sat_vec.len() < remaining_sat {
@@ -549,6 +804,10 @@ impl Property for ExtData {
             ops_count_sat: ops_count_sat
                 .map(|x: usize| (x + (n - 1) + 1 + (sum + ops_count_nsat_sum as i32) as usize)), //adds and equal
             ops_count_nsat: ops_count_nsat.map(|x| x + (n - 1) + 1), //adds and equal
+            stack_elem_count_sat,
+            stack_elem_count_dissat,
+            max_sat_size,
+            max_dissat_size,
             timelock_info: TimeLockInfo::combine_thresh_timelocks(k, timelocks),
         })
     }
