@@ -22,7 +22,6 @@ use std::convert::From;
 use std::marker::PhantomData;
 use std::{cmp, error, f64, fmt, mem};
 
-use miniscript::types::extra_props::MAX_OPS_PER_SCRIPT;
 use miniscript::types::{self, ErrorKind, ExtData, Property, Type};
 use miniscript::ScriptContext;
 use policy::Concrete;
@@ -54,24 +53,16 @@ pub enum CompilerError {
     TopLevelNonSafe,
     /// Non-Malleable compilation  does exists for the given sub-policy.
     ImpossibleNonMalleableCompilation,
-    /// Atleast one satisfaction path in the optimal Miniscript has opcodes
-    /// more than `MAX_OPS_PER_SCRIPT`(201). However, there may exist other
-    /// miniscripts which are under `MAX_OPS_PER_SCRIPT` but the compiler
-    /// currently does not find them.
-    MaxOpCountExceeded,
+    /// At least one satisfaction path in the optimal Miniscript has exceeded
+    /// the consensus or standardness limits.
+    /// There may exist other miniscripts which are under these limits but the
+    /// compiler currently does not find them.
+    LimitsExceeded,
     ///Policy related errors
     PolicyError(policy::concrete::PolicyError),
 }
 
-impl error::Error for CompilerError {
-    fn cause(&self) -> Option<&error::Error> {
-        None
-    }
-
-    fn description(&self) -> &str {
-        ""
-    }
-}
+impl error::Error for CompilerError {}
 
 impl fmt::Display for CompilerError {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
@@ -82,9 +73,8 @@ impl fmt::Display for CompilerError {
             CompilerError::ImpossibleNonMalleableCompilation => {
                 f.write_str("The compiler could not find any non-malleable compilation")
             }
-            CompilerError::MaxOpCountExceeded => f.write_str(
-                "Atleast one spending path has more op codes executed than \
-                 MAX_OPS_PER_SCRIPT",
+            CompilerError::LimitsExceeded => f.write_str(
+                "At least one spending path has exceeded the standardness or consensus limits",
             ),
             CompilerError::PolicyError(ref e) => fmt::Display::fmt(e, f),
         }
@@ -653,14 +643,12 @@ fn insert_elem<Pk: MiniscriptKey, Ctx: ScriptContext>(
 ) -> bool {
     // return malleable types directly. If a elem is malleable under current context,
     // all the casts to it are also going to be malleable
-    if !elem.ms.ty.mall.non_malleable && Ctx::check_frag_non_malleable(&elem.ms.node).is_ok() {
+    if !elem.ms.ty.mall.non_malleable && Ctx::check_terminal_non_malleable(&elem.ms.node).is_ok() {
         return false;
     }
 
-    if let Some(op_count) = elem.ms.ext.ops_count_sat {
-        if op_count > MAX_OPS_PER_SCRIPT {
-            return false;
-        }
+    if let Err(_) = Ctx::check_local_validity(&elem.ms) {
+        return false;
     }
 
     let elem_cost = elem.cost_1d(sat_prob, dissat_prob);
@@ -1027,12 +1015,12 @@ where
     }
     if ret.len() == 0 {
         // The only reason we are discarding elements out of compiler is because
-        // compilations exceed opcount or are non-malleable . If there no possible
-        // compilations for any policies regardless of dissat probability then it
-        // must have all compilations exceeded the Max Opcount because we already
-        // checked that policy must have non-malleable compilations before calling
-        // this compile function
-        Err(CompilerError::MaxOpCountExceeded)
+        // compilations exceeded consensus and standardness limits or are non-malleable.
+        // If there no possible compilations for any policies regardless of dissat
+        // probability then it must have all compilations exceeded consensus or standardness
+        // limits because we already checked that policy must have non-malleable compilations
+        // before calling this compile function
+        Err(CompilerError::LimitsExceeded)
     } else {
         policy_cache.insert((policy.clone(), ord_sat_prob, ord_dissat_prob), ret.clone());
         Ok(ret)
@@ -1140,7 +1128,7 @@ where
         })
         .map(|(_, val)| val)
         .min_by_key(|ext| OrdF64(ext.cost_1d(sat_prob, dissat_prob)))
-        .ok_or(CompilerError::MaxOpCountExceeded)
+        .ok_or(CompilerError::LimitsExceeded)
 }
 
 /// Obtain the <basic-type>.deu (e.g. W.deu, B.deu) expression with the given sat and dissat
@@ -1165,7 +1153,7 @@ where
         })
         .map(|(_, val)| val)
         .min_by_key(|ext| OrdF64(ext.cost_1d(sat_prob, dissat_prob)))
-        .ok_or(CompilerError::MaxOpCountExceeded)
+        .ok_or(CompilerError::LimitsExceeded)
 }
 
 #[cfg(test)]
@@ -1177,14 +1165,13 @@ mod tests {
     use std::str::FromStr;
     use std::string::String;
 
-    use miniscript::{satisfy, Segwitv0};
+    use miniscript::{satisfy, Legacy, Segwitv0};
     use policy::Liftable;
     use script_num_size;
     use BitcoinSig;
-    use DummyKey;
+    use NullCtx;
 
     type SPolicy = Concrete<String>;
-    type DummyPolicy = Concrete<DummyKey>;
     type BPolicy = Concrete<bitcoin::PublicKey>;
     type DummySegwitAstElemExt = policy::compiler::AstElemExt<String, Segwitv0>;
     type SegwitMiniScript = Miniscript<bitcoin::PublicKey, Segwitv0>;
@@ -1215,8 +1202,8 @@ mod tests {
     }
 
     fn policy_compile_lift_check(s: &str) -> Result<(), CompilerError> {
-        let policy = DummyPolicy::from_str(s).expect("parse");
-        let miniscript: Miniscript<DummyKey, Segwitv0> = policy.compile()?;
+        let policy = SPolicy::from_str(s).expect("parse");
+        let miniscript: Miniscript<String, Segwitv0> = policy.compile()?;
 
         assert_eq!(
             policy.lift().unwrap().sorted(),
@@ -1228,21 +1215,21 @@ mod tests {
     #[test]
     fn compile_timelocks() {
         // artificially create a policy that is problematic and try to compile
-        let pol: DummyPolicy = Concrete::And(vec![
-            Concrete::Key(DummyKey),
+        let pol: SPolicy = Concrete::And(vec![
+            Concrete::Key("A".to_string()),
             Concrete::And(vec![Concrete::After(9), Concrete::After(1000_000_000)]),
         ]);
         assert!(pol.compile::<Segwitv0>().is_err());
 
         // This should compile
-        let pol: DummyPolicy =
-            DummyPolicy::from_str("and(pk(),or(and(after(9),pk()),and(after(1000000000),pk())))")
+        let pol: SPolicy =
+            SPolicy::from_str("and(pk(A),or(and(after(9),pk(B)),and(after(1000000000),pk(C))))")
                 .unwrap();
         assert!(pol.compile::<Segwitv0>().is_ok());
     }
     #[test]
     fn compile_basic() {
-        assert!(policy_compile_lift_check("pk()").is_ok());
+        assert!(policy_compile_lift_check("pk(A)").is_ok());
         assert_eq!(
             policy_compile_lift_check("after(9)"),
             Err(CompilerError::TopLevelNonSafe)
@@ -1257,24 +1244,24 @@ mod tests {
             ),
             Err(CompilerError::TopLevelNonSafe)
         );
-        assert!(policy_compile_lift_check("and(pk(),pk())").is_ok());
-        assert!(policy_compile_lift_check("or(pk(),pk())").is_ok());
-        assert!(policy_compile_lift_check("thresh(2,pk(),pk(),pk())").is_ok());
+        assert!(policy_compile_lift_check("and(pk(A),pk(B))").is_ok());
+        assert!(policy_compile_lift_check("or(pk(A),pk(B))").is_ok());
+        assert!(policy_compile_lift_check("thresh(2,pk(A),pk(B),pk(C))").is_ok());
 
         assert_eq!(
-            policy_compile_lift_check("thresh(2,after(9),after(9),pk())"),
+            policy_compile_lift_check("thresh(2,after(9),after(9),pk(A))"),
             Err(CompilerError::TopLevelNonSafe)
         );
 
         assert_eq!(
-            policy_compile_lift_check("and(pk(),or(after(9),after(9)))"),
+            policy_compile_lift_check("and(pk(A),or(after(9),after(9)))"),
             Err(CompilerError::ImpossibleNonMalleableCompilation)
         );
     }
 
     #[test]
     fn compile_q() {
-        let policy = SPolicy::from_str("or(1@and(pk(),pk()),127@pk())").expect("parsing");
+        let policy = SPolicy::from_str("or(1@and(pk(A),pk(B)),127@pk(C))").expect("parsing");
         let compilation: DummySegwitAstElemExt =
             best_t(&mut BTreeMap::new(), &policy, 1.0, None).unwrap();
 
@@ -1285,7 +1272,7 @@ mod tests {
         );
 
         let policy = SPolicy::from_str(
-                "and(and(and(or(127@thresh(2,pk(),pk(),thresh(2,or(127@pk(),1@pk()),after(100),or(and(pk(),after(200)),and(pk(),sha256(66687aadf862bd776c8fc18b8e9f8e20089714856ee233b3902a591d0d5f2925))),pk())),1@pk()),sha256(66687aadf862bd776c8fc18b8e9f8e20089714856ee233b3902a591d0d5f2925)),or(127@pk(),1@after(300))),or(127@after(400),pk()))"
+                "and(and(and(or(127@thresh(2,pk(A),pk(B),thresh(2,or(127@pk(A),1@pk(B)),after(100),or(and(pk(C),after(200)),and(pk(D),sha256(66687aadf862bd776c8fc18b8e9f8e20089714856ee233b3902a591d0d5f2925))),pk(E))),1@pk(F)),sha256(66687aadf862bd776c8fc18b8e9f8e20089714856ee233b3902a591d0d5f2925)),or(127@pk(G),1@after(300))),or(127@after(400),pk(H)))"
             ).expect("parsing");
         let compilation: DummySegwitAstElemExt =
             best_t(&mut BTreeMap::new(), &policy, 1.0, None).unwrap();
@@ -1305,7 +1292,7 @@ mod tests {
         let policy: BPolicy = Concrete::Key(keys[0].clone());
         let ms: SegwitMiniScript = policy.compile().unwrap();
         assert_eq!(
-            ms.encode(),
+            ms.encode(NullCtx),
             script::Builder::new()
                 .push_key(&keys[0])
                 .push_opcode(opcodes::all::OP_CHECKSIG)
@@ -1321,7 +1308,7 @@ mod tests {
         );
         let ms: SegwitMiniScript = policy.compile().unwrap();
         assert_eq!(
-            ms.encode(),
+            ms.encode(NullCtx),
             script::Builder::new()
                 .push_opcode(opcodes::all::OP_PUSHNUM_2)
                 .push_key(&keys[5])
@@ -1393,14 +1380,19 @@ mod tests {
             right_sat.insert(keys[i].to_pubkeyhash(), (keys[i], bitcoinsig));
         }
 
-        assert!(ms.satisfy(no_sat).is_none());
-        assert!(ms.satisfy(&left_sat).is_some());
-        assert!(ms.satisfy((&right_sat, satisfy::Older(10001))).is_some());
+        assert!(ms.satisfy(no_sat, NullCtx).is_err());
+        assert!(ms.satisfy(&left_sat, NullCtx).is_ok());
+        assert!(ms
+            .satisfy((&right_sat, satisfy::Older(10001)), NullCtx)
+            .is_ok());
         //timelock not met
-        assert!(ms.satisfy((&right_sat, satisfy::Older(9999))).is_none());
+        assert!(ms
+            .satisfy((&right_sat, satisfy::Older(9999)), NullCtx)
+            .is_err());
 
         assert_eq!(
-            ms.satisfy((left_sat, satisfy::Older(9999))).unwrap(),
+            ms.satisfy((left_sat, satisfy::Older(9999)), NullCtx)
+                .unwrap(),
             vec![
                 // sat for left branch
                 vec![],
@@ -1411,7 +1403,8 @@ mod tests {
         );
 
         assert_eq!(
-            ms.satisfy((right_sat, satisfy::Older(10000))).unwrap(),
+            ms.satisfy((right_sat, satisfy::Older(10000)), NullCtx)
+                .unwrap(),
             vec![
                 // sat for right branch
                 vec![],
@@ -1458,11 +1451,14 @@ mod tests {
             let big_thresh_ms: SegwitMiniScript = big_thresh.compile().unwrap();
             if *k == 21 {
                 // N * (PUSH + pubkey + CHECKSIGVERIFY)
-                assert_eq!(big_thresh_ms.script_size(), keys.len() * (1 + 33 + 1));
+                assert_eq!(
+                    big_thresh_ms.script_size(NullCtx),
+                    keys.len() * (1 + 33 + 1)
+                );
             } else {
                 // N * (PUSH + pubkey + CHECKSIG + ADD + SWAP) + N EQUAL
                 assert_eq!(
-                    big_thresh_ms.script_size(),
+                    big_thresh_ms.script_size(NullCtx),
                     keys.len() * (1 + 33 + 3) + script_num_size(*k) + 1 - 2 // minus one SWAP and one ADD
                 );
                 let big_thresh_ms_expected = ms_str!(
@@ -1473,6 +1469,99 @@ mod tests {
             };
         }
     }
+
+    #[test]
+    fn segwit_limits() {
+        // Hit the maximum witness script size limit.
+        // or(thresh(52, [pubkey; 52]), thresh(52, [pubkey; 52])) results in a 3642-bytes long
+        // witness script with only 54 stack elements
+        let (keys, _) = pubkeys_and_a_sig(104);
+        let keys_a: Vec<Concrete<bitcoin::PublicKey>> = keys[..keys.len() / 2]
+            .iter()
+            .map(|pubkey| Concrete::Key(*pubkey))
+            .collect();
+        let keys_b: Vec<Concrete<bitcoin::PublicKey>> = keys[keys.len() / 2..]
+            .iter()
+            .map(|pubkey| Concrete::Key(*pubkey))
+            .collect();
+
+        let thresh_res: Result<SegwitMiniScript, _> = Concrete::Or(vec![
+            (1, Concrete::Threshold(keys_a.len(), keys_a)),
+            (1, Concrete::Threshold(keys_b.len(), keys_b)),
+        ])
+        .compile();
+        let script_size = thresh_res.clone().and_then(|m| Ok(m.script_size(NullCtx)));
+        assert_eq!(
+            thresh_res,
+            Err(CompilerError::LimitsExceeded),
+            "Compilation succeeded with a witscript size of '{:?}'",
+            script_size,
+        );
+
+        // Hit the maximum witness stack elements limit
+        let (keys, _) = pubkeys_and_a_sig(100);
+        let keys: Vec<Concrete<bitcoin::PublicKey>> =
+            keys.iter().map(|pubkey| Concrete::Key(*pubkey)).collect();
+        let thresh_res: Result<SegwitMiniScript, _> =
+            Concrete::Threshold(keys.len(), keys).compile();
+        let n_elements = thresh_res
+            .clone()
+            .and_then(|m| Ok(m.max_satisfaction_witness_elements()));
+        assert_eq!(
+            thresh_res,
+            Err(CompilerError::LimitsExceeded),
+            "Compilation succeeded with '{:?}' elements",
+            n_elements,
+        );
+    }
+
+    #[test]
+    fn shared_limits() {
+        // Test the maximum number of OPs with a 67-of-68 multisig
+        let (keys, _) = pubkeys_and_a_sig(68);
+        let keys: Vec<Concrete<bitcoin::PublicKey>> =
+            keys.iter().map(|pubkey| Concrete::Key(*pubkey)).collect();
+        let thresh_res: Result<SegwitMiniScript, _> =
+            Concrete::Threshold(keys.len() - 1, keys).compile();
+        let ops_count = thresh_res.clone().and_then(|m| Ok(m.ext.ops_count_sat));
+        assert_eq!(
+            thresh_res,
+            Err(CompilerError::LimitsExceeded),
+            "Compilation succeeded with '{:?}' OP count (sat)",
+            ops_count,
+        );
+        // For legacy too..
+        let (keys, _) = pubkeys_and_a_sig(68);
+        let keys: Vec<Concrete<bitcoin::PublicKey>> =
+            keys.iter().map(|pubkey| Concrete::Key(*pubkey)).collect();
+        let thresh_res = Concrete::Threshold(keys.len() - 1, keys).compile::<Legacy>();
+        let ops_count = thresh_res.clone().and_then(|m| Ok(m.ext.ops_count_sat));
+        assert_eq!(
+            thresh_res,
+            Err(CompilerError::LimitsExceeded),
+            "Compilation succeeded with '{:?}' OP count (sat)",
+            ops_count,
+        );
+
+        // Test that we refuse to compile policies with duplicated keys
+        let (keys, _) = pubkeys_and_a_sig(1);
+        let key = Concrete::Key(keys[0]);
+        let res = Concrete::Or(vec![(1, key.clone()), (1, key.clone())]).compile::<Segwitv0>();
+        assert_eq!(
+            res,
+            Err(CompilerError::PolicyError(
+                policy::concrete::PolicyError::DuplicatePubKeys
+            ))
+        );
+        // Same for legacy
+        let res = Concrete::Or(vec![(1, key.clone()), (1, key)]).compile::<Legacy>();
+        assert_eq!(
+            res,
+            Err(CompilerError::PolicyError(
+                policy::concrete::PolicyError::DuplicatePubKeys
+            ))
+        );
+    }
 }
 
 #[cfg(all(test, feature = "unstable"))]
@@ -1482,19 +1571,18 @@ mod benches {
 
     use super::{CompilerError, Concrete};
     use miniscript::Segwitv0;
-    use DummyKey;
     use Miniscript;
-    type DummySegwitMiniscriptRes = Result<Miniscript<DummyKey, Segwitv0>, CompilerError>;
+    type SegwitMsRes = Result<Miniscript<String, Segwitv0>, CompilerError>;
     #[bench]
     pub fn compile_basic(bh: &mut Bencher) {
         let h = (0..64).map(|_| "a").collect::<String>();
-        let pol = Concrete::<DummyKey>::from_str(&format!(
-            "and(thresh(2,and(sha256({}),or(sha256({}),pk())),pk(),pk(),pk(),sha256({})),pk())",
+        let pol = Concrete::<String>::from_str(&format!(
+            "and(thresh(2,and(sha256({}),or(sha256({}),pk(A))),pk(B),pk(C),pk(D),sha256({})),pk(E))",
             h, h, h
         ))
         .expect("parsing");
         bh.iter(|| {
-            let pt: DummySegwitMiniscriptRes = pol.compile();
+            let pt: SegwitMsRes = pol.compile();
             black_box(pt).unwrap();
         });
     }
@@ -1502,22 +1590,22 @@ mod benches {
     #[bench]
     pub fn compile_large(bh: &mut Bencher) {
         let h = (0..64).map(|_| "a").collect::<String>();
-        let pol = Concrete::<DummyKey>::from_str(
-            &format!("or(pk(),thresh(9,sha256({}),pk(),pk(),and(or(pk(),pk()),pk()),after(100),pk(),pk(),pk(),pk(),and(pk(),pk())))", h)
+        let pol = Concrete::<String>::from_str(
+            &format!("or(pk(L),thresh(9,sha256({}),pk(A),pk(B),and(or(pk(C),pk(D)),pk(E)),after(100),pk(F),pk(G),pk(H),pk(I),and(pk(J),pk(K))))", h)
         ).expect("parsing");
         bh.iter(|| {
-            let pt: DummySegwitMiniscriptRes = pol.compile();
+            let pt: SegwitMsRes = pol.compile();
             black_box(pt).unwrap();
         });
     }
 
     #[bench]
     pub fn compile_xlarge(bh: &mut Bencher) {
-        let pol = Concrete::<DummyKey>::from_str(
-            "or(pk(),thresh(4,pk(),older(100),pk(),and(after(100),or(pk(),or(pk(),and(pk(),thresh(2,pk(),or(pk(),and(thresh(5,pk(),or(pk(),pk()),pk(),pk(),pk(),pk(),pk(),pk(),pk(),pk(),pk()),pk())),pk(),or(and(pk(),pk()),pk()),after(100)))))),pk()))"
+        let pol = Concrete::<String>::from_str(
+            "or(pk(A),thresh(4,pk(B),older(100),pk(C),and(after(100),or(pk(D),or(pk(E),and(pk(F),thresh(2,pk(G),or(pk(H),and(thresh(5,pk(I),or(pk(J),pk(K)),pk(L),pk(M),pk(N),pk(O),pk(P),pk(Q),pk(R),pk(S),pk(T)),pk(U))),pk(V),or(and(pk(W),pk(X)),pk(Y)),after(100)))))),pk(Z)))"
         ).expect("parsing");
         bh.iter(|| {
-            let pt: DummySegwitMiniscriptRes = pol.compile();
+            let pt: SegwitMsRes = pol.compile();
             black_box(pt).unwrap();
         });
     }
