@@ -17,12 +17,14 @@
 
 use bitcoin::hashes::hex::FromHex;
 use bitcoin::hashes::{hash160, ripemd160, sha256, sha256d};
+use std::collections::HashSet;
 use std::{error, fmt, str};
 
 use super::ENTAILMENT_MAX_TERMINALS;
 use errstr;
 use expression::{self, FromTree};
-use miniscript::types::extra_props::{TimeLockInfo, HEIGHT_TIME_THRESHOLD};
+use miniscript::limits::{HEIGHT_TIME_THRESHOLD, SEQUENCE_LOCKTIME_TYPE_FLAG};
+use miniscript::types::extra_props::TimeLockInfo;
 #[cfg(feature = "compiler")]
 use miniscript::ScriptContext;
 #[cfg(feature = "compiler")]
@@ -86,16 +88,11 @@ pub enum PolicyError {
     /// lifting error: Cannot lift policies that have
     /// a combination of height and timelocks.
     HeightTimeLockCombination,
+    /// Duplicate Public Keys
+    DuplicatePubKeys,
 }
 
-impl error::Error for PolicyError {
-    fn description(&self) -> &str {
-        ""
-    }
-    fn cause(&self) -> Option<&error::Error> {
-        None
-    }
-}
+impl error::Error for PolicyError {}
 
 impl fmt::Display for PolicyError {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
@@ -112,10 +109,10 @@ impl fmt::Display for PolicyError {
             }
             PolicyError::ZeroTime => f.write_str("Time must be greater than 0; n > 0"),
             PolicyError::InsufficientArgsforAnd => {
-                f.write_str("Semantic Policy 'And' fragment must have atleast 2 args ")
+                f.write_str("Semantic Policy 'And' fragment must have at least 2 args ")
             }
             PolicyError::InsufficientArgsforOr => {
-                f.write_str("Semantic Policy 'Or' fragment must have atleast 2 args ")
+                f.write_str("Semantic Policy 'Or' fragment must have at least 2 args ")
             }
             PolicyError::EntailmentMaxTerminals => write!(
                 f,
@@ -125,6 +122,7 @@ impl fmt::Display for PolicyError {
             PolicyError::HeightTimeLockCombination => {
                 f.write_str("Cannot lift policies that have a heightlock and timelock combination")
             }
+            PolicyError::DuplicatePubKeys => f.write_str("Policy contains duplicate keys"),
         }
     }
 }
@@ -180,9 +178,46 @@ impl<Pk: MiniscriptKey> Policy<Pk> {
         }
     }
 
+    /// Get all keys in the policy
+    pub fn keys(&self) -> Vec<&Pk> {
+        match *self {
+            Policy::Key(ref pk) => vec![pk],
+            Policy::Threshold(_k, ref subs) => subs
+                .iter()
+                .map(|sub| sub.keys())
+                .flatten()
+                .collect::<Vec<_>>(),
+            Policy::And(ref subs) => subs
+                .iter()
+                .map(|sub| sub.keys())
+                .flatten()
+                .collect::<Vec<_>>(),
+            Policy::Or(ref subs) => subs
+                .iter()
+                .map(|(ref _k, ref sub)| sub.keys())
+                .flatten()
+                .collect::<Vec<_>>(),
+            // map all hashes and time
+            _ => vec![],
+        }
+    }
+
+    /// Check whether the policy contains duplicate public keys
+    pub fn check_duplicate_keys(&self) -> Result<(), PolicyError> {
+        let pks = self.keys();
+        let pks_len = pks.len();
+        let unique_pks_len = pks.into_iter().collect::<HashSet<_>>().len();
+
+        if pks_len > unique_pks_len {
+            Err(PolicyError::DuplicatePubKeys)
+        } else {
+            Ok(())
+        }
+    }
+
     /// Checks whether the given concrete policy contains a combination of
     /// timelocks and heightlocks.
-    /// Returns an error if there is atleast one satisfaction that contains
+    /// Returns an error if there is at least one satisfaction that contains
     /// a combination of hieghtlock and timelock.
     pub fn check_timelocks(&self) -> Result<(), PolicyError> {
         let timelocks = self.check_timelocks_helper();
@@ -213,8 +248,8 @@ impl<Pk: MiniscriptKey> Policy<Pk> {
                 contains_combination: false,
             },
             Policy::Older(t) => TimeLockInfo {
-                csv_with_height: t < HEIGHT_TIME_THRESHOLD,
-                csv_with_time: t >= HEIGHT_TIME_THRESHOLD,
+                csv_with_height: (t & SEQUENCE_LOCKTIME_TYPE_FLAG) == 0,
+                csv_with_time: (t & SEQUENCE_LOCKTIME_TYPE_FLAG) != 0,
                 cltv_with_height: false,
                 cltv_with_time: false,
                 contains_combination: false,
@@ -242,6 +277,7 @@ impl<Pk: MiniscriptKey> Policy<Pk> {
     /// combination of timelocks and heightlocks
     pub fn is_valid(&self) -> Result<(), PolicyError> {
         self.check_timelocks()?;
+        self.check_duplicate_keys()?;
         match *self {
             Policy::And(ref subs) => {
                 if subs.len() != 2 {
